# Generated by Django 4.2.16 on 2024-10-09 09:41

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


def assign_contributor(apps, schema_editor):
    """Assign initial contributor to annotation values."""

    AnnotationValue = apps.get_model("flow", "AnnotationValue")
    for annotation_value in AnnotationValue.objects.iterator():
        annotation_value.contributor = annotation_value.entity.contributor
        annotation_value.save(update_fields=["contributor"])


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("flow", "0029_remove_annotationfield_uniquetogether_name_group_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="annotationvalue",
            name="contributor",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.RunPython(assign_contributor),
        migrations.AlterField(
            model_name="annotationvalue",
            name="contributor",
            field=models.ForeignKey(
                default=None,
                on_delete=django.db.models.deletion.PROTECT,
                to=settings.AUTH_USER_MODEL,
            ),
            preserve_default=False,
        ),
    ]



"""ORM signal handlers."""

from typing import Union

from django import dispatch
from django.db.models import Model
from django.db.models import signals as model_signals

from resolwe.flow.models import Collection, Data, Entity
from resolwe.flow.signals import post_duplicate
from resolwe.permissions.models import Permission

from .models import Observable, Observer
from .protocol import (
    ChangeType,
    post_container_changed,
    post_permission_changed,
    pre_container_changed,
    pre_permission_changed,
    suppress_notifications_attribute,
)

# Global 'in migrations' flag to ignore signals during migrations.
# Signal handlers that access the database can crash the migration process.
IN_MIGRATIONS = False

Duplicate = Union[Data, Entity, Collection]


@dispatch.receiver(model_signals.pre_migrate)
def model_pre_migrate(*args, **kwargs):
    """Set 'in migrations' flag."""
    global IN_MIGRATIONS
    IN_MIGRATIONS = True


@dispatch.receiver(model_signals.post_migrate)
def model_post_migrate(*args, **kwargs):
    """Clear 'in migrations' flag."""
    global IN_MIGRATIONS
    IN_MIGRATIONS = False


def skip_in_migrations(function):
    """Execute the given method if not in migrations."""

    def wrapped(*args, **kwargs):
        if not IN_MIGRATIONS:
            function(*args, **kwargs)

    return wrapped


@dispatch.receiver(pre_permission_changed)
@skip_in_migrations
def prepare_permission_change(instance, **kwargs):
    """Store old permissions for an object whose permissions are about to change."""
    instance._old_viewers = instance.users_with_permission(
        Permission.VIEW, with_superusers=True
    )


@dispatch.receiver(pre_container_changed)
@skip_in_migrations
def prepare_container_change(instance, **kwargs):
    """Store old containerds for an object."""
    prepare_permission_change(instance)
    instance._old_containers = set(instance.containers)


@dispatch.receiver(post_container_changed)
@skip_in_migrations
def handle_container_change(instance, **kwargs):
    """Send notifications when object is moved to another container."""
    old_containers = getattr(instance, "_old_containers", set())
    new_containers = set(instance.containers)
    removed_from = old_containers - new_containers
    added_to = new_containers - old_containers

    # Handle possible permission change (CREATE/DELETE) on the object itself. Do
    # not send notifications on the containers, this is handled bellow.
    handle_permission_change(instance, observe_containers=False)
    # Handle notifications (CREATE/DELETE) on the containers.
    Observer.observe_instance_container(instance, ChangeType.DELETE, removed_from)
    Observer.observe_instance_container(instance, ChangeType.CREATE, added_to)
    # Handle UPDATE notifications on the instance (container property has changed).
    Observer.observe_instance_changes(instance, ChangeType.UPDATE)


@dispatch.receiver(post_permission_changed)
@skip_in_migrations
def handle_permission_change(instance, **kwargs):
    """Compare permissions for an object whose permissions changed."""
    new = set(instance.users_with_permission(Permission.VIEW, with_superusers=True))
    # The "_old_viewers" property may not exist. For instance if data object is
    # created in the collection and collection permissions are assigned to it
    # without set_permission call.
    old = set(getattr(instance, "_old_viewers", []))
    gains = new - old
    losses = old - new
    observe_containers = kwargs.get("observe_containers", True)
    Observer.observe_permission_changes(instance, gains, losses, observe_containers)


@dispatch.receiver(model_signals.post_save)
@skip_in_migrations
def observe_model_modification(
    sender: type, instance: Model, created: bool = False, **kwargs
):
    """Receive model updates.

    We have to consider the following options.
    1. The object was created: send CREATE notifications to users with VIEW permission.
    2. The object was changed: send UPDATE notifications to users with VIEW permission.

    Do not send notifications to the containers when object is moving between them.
    """
    if isinstance(instance, Observable):
        if created:
            handle_permission_change(instance)
        elif not getattr(instance, suppress_notifications_attribute, False):
            Observer.observe_instance_changes(instance, ChangeType.UPDATE)
            Observer.observe_instance_container(instance, ChangeType.UPDATE)


@dispatch.receiver(model_signals.pre_delete)
@skip_in_migrations
def observe_model_deletion(sender: type, instance: Model, **kwargs):
    """Receive model deletions."""
    if isinstance(instance, Observable):
        Observer.observe_instance_changes(instance, ChangeType.DELETE)
        Observer.observe_instance_container(instance, ChangeType.DELETE)


@dispatch.receiver(post_duplicate)
def post_duplicate_models(
    sender, instances: list[Duplicate], old_instances: list[Duplicate], **kwargs
):
    """Send create notification when duplicates are created.

    This is necessary since bulk_duplicate does not trigger post_save signals.
    """
    # Make sure all instances have the same permisions.
    gains = dict()
    for instance in instances:
        if instance.permission_group_id not in gains:
            gains[instance.permission_group_id] = set(
                instance.users_with_permission(Permission.VIEW, with_superusers=True)
            )

    for instance in instances:
        Observer.observe_permission_changes(
            instance, gains[instance.permission_group_id], set()
        )



# Generated by Django 4.2.16 on 2024-10-23 10:27

import versionfield.fields
from django.db import migrations, models

import resolwe.flow.models.annotations
import resolwe.flow.models.base
import resolwe.flow.models.fields


def assign_modify_to_created(apps, schema_editor):
    """Assign created date to annotation values."""

    AnnotationValue = apps.get_model("flow", "AnnotationValue")
    for annotation_value in AnnotationValue.objects.iterator():
        annotation_value.created = annotation_value.modified
        annotation_value.save(update_fields=["created"])


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0030_annotationvalue_contributor"),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name="annotationvalue",
            name="uniquetogether_entity_field",
        ),
        migrations.AddField(
            model_name="annotationvalue",
            name="created",
            field=models.DateTimeField(auto_now_add=True, db_index=True, null=None),
            preserve_default=False,
        ),
        migrations.RunPython(assign_modify_to_created),
        migrations.AlterField(
            model_name="annotationvalue",
            name="created",
            field=models.DateTimeField(auto_now_add=True, db_index=True, default=None),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="annotationvalue",
            name="name",
            field=models.CharField(default="", max_length=100),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="annotationvalue",
            name="slug",
            field=resolwe.flow.models.fields.ResolweSlugField(
                default="",
                max_length=100,
                populate_from=resolwe.flow.models.annotations._slug_for_annotation_value,
                unique_with=("version",),
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="annotationvalue",
            name="version",
            field=versionfield.fields.VersionField(default="0.0.0"),
        ),
        migrations.AlterField(
            model_name="annotationvalue",
            name="_value",
            field=models.JSONField(default=dict, null=True),
        ),
        migrations.AlterField(
            model_name="annotationvalue",
            name="modified",
            field=resolwe.flow.models.base.ModifiedField(auto_now=True, db_index=True),
        ),
        migrations.AddConstraint(
            model_name="annotationvalue",
            constraint=models.UniqueConstraint(
                fields=("entity", "field", "created"),
                name="uniquetogether_entity_field_created",
            ),
        ),
    ]



# Generated by Django 4.2.6 on 2023-10-27 07:52

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("flow", "0017_collectionhistory_datahistory_sizechange_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="annotationvalue",
            options={"ordering": ["field__group__sort_order", "field__sort_order"]},
        ),
    ]



# Generated by Django 4.2.11 on 2024-05-27 11:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0026_create_unaccent_full_text_search_config"),
    ]

    operations = [
        migrations.AddField(
            model_name="worker",
            name="private_key",
            field=models.BinaryField(default=b"", max_length=40),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="worker",
            name="public_key",
            field=models.BinaryField(default=b"", max_length=40),
            preserve_default=False,
        ),
    ]



# Generated by Django 4.2.2 on 2023-10-05 11:10

import django.contrib.postgres.fields.ranges
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("flow", "0016_annotationvalue_uniquetogether_entity_field"),
    ]

    operations = [
        migrations.CreateModel(
            name="CollectionHistory",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("valid", django.contrib.postgres.fields.ranges.DateTimeRangeField()),
                ("deleted", models.DateTimeField(blank=True, null=True)),
                (
                    "datum",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="history",
                        to="flow.collection",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="DataHistory",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("valid", django.contrib.postgres.fields.ranges.DateTimeRangeField()),
                ("deleted", models.DateTimeField(blank=True, null=True)),
                (
                    "datum",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="history",
                        to="flow.data",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="SizeChange",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("timestamp", models.DateTimeField(auto_now_add=True)),
                ("value", models.BigIntegerField()),
                (
                    "history",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="%(class)s",
                        to="flow.datahistory",
                    ),
                ),
            ],
            options={
                "ordering": ["timestamp"],
                "abstract": False,
                "indexes": [
                    models.Index(fields=["timestamp"], name="idx_sizechange_time")
                ],
            },
        ),
        migrations.CreateModel(
            name="ProcessingHistory",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("timestamp", models.DateTimeField(auto_now_add=True)),
                (
                    "interval",
                    django.contrib.postgres.fields.ranges.DateTimeRangeField(),
                ),
                ("allocated_cpu", models.IntegerField()),
                ("allocated_memory", models.BigIntegerField()),
                ("node_cpu", models.IntegerField()),
                ("node_memory", models.BigIntegerField()),
                (
                    "history",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="%(class)s",
                        to="flow.datahistory",
                    ),
                ),
            ],
            options={
                "ordering": ["timestamp"],
                "abstract": False,
                "indexes": [
                    models.Index(
                        fields=["timestamp"], name="idx_processinghistory_time"
                    )
                ],
            },
        ),
        migrations.CreateModel(
            name="DataSlugChange",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("timestamp", models.DateTimeField(auto_now_add=True)),
                ("value", models.CharField(max_length=100)),
                (
                    "history",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="%(class)s",
                        to="flow.datahistory",
                    ),
                ),
            ],
            options={
                "ordering": ["timestamp"],
                "abstract": False,
                "indexes": [
                    models.Index(fields=["timestamp"], name="idx_dataslugchange_time")
                ],
            },
        ),
        migrations.CreateModel(
            name="DataNameChange",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("timestamp", models.DateTimeField(auto_now_add=True)),
                ("value", models.CharField(max_length=100)),
                (
                    "history",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="%(class)s",
                        to="flow.datahistory",
                    ),
                ),
            ],
            options={
                "ordering": ["timestamp"],
                "abstract": False,
                "indexes": [
                    models.Index(fields=["timestamp"], name="idx_datanamechange_time")
                ],
            },
        ),
        migrations.CreateModel(
            name="CollectionSlugChange",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("timestamp", models.DateTimeField(auto_now_add=True)),
                ("value", models.CharField(max_length=100)),
                (
                    "history",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="%(class)s",
                        to="flow.collectionhistory",
                    ),
                ),
            ],
            options={
                "ordering": ["timestamp"],
                "abstract": False,
                "indexes": [
                    models.Index(
                        fields=["timestamp"], name="idx_collectionslugchange_time"
                    )
                ],
            },
        ),
        migrations.CreateModel(
            name="CollectionNameChange",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("timestamp", models.DateTimeField(auto_now_add=True)),
                ("value", models.CharField(max_length=100)),
                (
                    "history",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="%(class)s",
                        to="flow.collectionhistory",
                    ),
                ),
            ],
            options={
                "ordering": ["timestamp"],
                "abstract": False,
                "indexes": [
                    models.Index(
                        fields=["timestamp"], name="idx_collectionnamechange_time"
                    )
                ],
            },
        ),
        migrations.CreateModel(
            name="CollectionChange",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("timestamp", models.DateTimeField(auto_now_add=True)),
                (
                    "history",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="%(class)s",
                        to="flow.datahistory",
                    ),
                ),
                (
                    "value",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="collection_changes",
                        to="flow.collectionhistory",
                    ),
                ),
            ],
            options={
                "ordering": ["timestamp"],
                "abstract": False,
                "indexes": [
                    models.Index(fields=["timestamp"], name="idx_collectionchange_time")
                ],
            },
        ),
    ]



# Generated by Django 4.2 on 2023-05-10 11:41

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("flow", "0011_annotationpreset_contributor_correct_user"),
    ]

    operations = [
        migrations.AlterField(
            model_name="relation",
            name="category",
            field=models.CharField(max_length=100),
        ),
    ]



""".. Ignore pydocstyle D400.

======================
Permissions Management
======================

.. automodule:: resolwe.flow.management.commands.register
    :members:

"""

from django.conf import settings
from django.contrib.auth import get_user_model
from django.db.models import signals


def create_anonymous_user(sender, **kwargs):
    """Create anonymous User instance if it does not exist."""
    User = get_user_model()
    try:
        User.objects.get(username=settings.ANONYMOUS_USER_NAME)
    except User.DoesNotExist:
        User.objects.create_user(username=settings.ANONYMOUS_USER_NAME)


# Create Anonymous user on post migrate signal. This includes "flush" which is
# performed it tests originating from TransactionTestCase.
signals.post_migrate.connect(create_anonymous_user)



# Generated by Django 3.1.7 on 2021-10-12 10:39
import os

from django.db import connection, migrations


def load_triggers(apps, schema_editor):
    file_names = [
        "triggers_entity.sql",
        "triggers_collection.sql",
        "triggers_data.sql",
        "utils.sql",
    ]
    with connection.cursor() as c:
        for file_name in file_names:
            file_path = os.path.join(os.path.dirname(__file__), file_name)
            with open(file_path) as fh:
                sql_statement = fh.read()
            c.execute(sql_statement)


class Migration(migrations.Migration):
    dependencies = [("flow", "0027_worker_private_key_worker_public_key")]

    operations = [
        migrations.RunPython(load_triggers),
        migrations.RunSQL("UPDATE flow_collection SET id=id;"),
        migrations.RunSQL("UPDATE flow_entity SET id=id;"),
        migrations.RunSQL("UPDATE flow_data SET id=id;"),
    ]



# pylint: disable=missing-docstring
import shutil
import unittest
from datetime import timedelta

from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.test import TransactionTestCase as DjangoTransactionTestCase
from django.urls import reverse
from django.utils.timezone import now
from rest_framework import exceptions, status
from rest_framework.test import APIClient

from resolwe.flow.models import Collection, Data
from resolwe.flow.serializers import ContributorSerializer
from resolwe.flow.views import DataViewSet
from resolwe.observers.models import BackgroundTask
from resolwe.permissions.models import Permission, get_anonymous_user
from resolwe.test import ResolweAPITestCase, TransactionResolweAPITestCase

DATE_FORMAT = r"%Y-%m-%dT%H:%M:%S.%f"

MESSAGES = {
    "NOT_FOUND": "No Data matches the given query.",
    "ONE_ID_REQUIRED": "Exactly one id required on create.",
}


class DataTestCaseCommonMixin:
    fixtures = [
        "users.yaml",
        "permissions.yaml",
        "collections.yaml",
        "processes.yaml",
        "data.yaml",
    ]

    def setUp(self):
        self.data1 = Data.objects.get(pk=1)
        self.data2 = Data.objects.get(pk=2)

        self.resource_name = "data"
        self.viewset = DataViewSet

        self.data = {
            "name": "New data",
            "slug": "new_data",
            "collection": {"id": 1},
            "process": {"slug": "test_process"},
        }

        self.client = APIClient()
        super().setUp()

    def tearDown(self):
        for data in Data.objects.all():
            if data.location:
                shutil.rmtree(data.location.get_path(), ignore_errors=True)

        super().tearDown()


class DataTestCase(DataTestCaseCommonMixin, ResolweAPITestCase):
    def test_get_list(self):
        resp = self._get_list(self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(len(resp.data), 2)

    def test_get_list_public_user(self):
        resp = self._get_list()
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(len(resp.data), 1)

    def test_get_list_admin(self):
        resp = self._get_list(self.admin)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(len(resp.data), 2)

    @unittest.skipIf(
        True,
        "since PR308: this test uses transactions, incompatible with the separated manager",
    )
    def test_post(self):
        # logged-in user w/ perms
        collection_n = Data.objects.count()

        resp = self._post(self.data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)
        self.assertEqual(Data.objects.count(), collection_n + 1)

        d = Data.objects.get(pk=resp.data["id"])
        self.assertTrue(now() - d.modified < timedelta(seconds=1))
        self.assertTrue(now() - d.created < timedelta(seconds=1))
        self.assertEqual(d.status, "OK")

        self.assertTrue(now() - d.started < timedelta(seconds=1))
        self.assertTrue(now() - d.finished < timedelta(seconds=1))
        self.assertEqual(d.contributor_id, self.user1.pk)

    def test_post_invalid_fields(self):
        data_n = Data.objects.count()

        self.data["collection"] = {"id": 42}
        resp = self._post(self.data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            resp.data["collection"][0],
            "Invalid collection value: {'id': 42} - object does not exist.",
        )

        self.data["collection"] = {"id": 1}
        self.data["process"] = {"id": 42}
        resp = self._post(self.data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            str(resp.data["process"][0]),
            "Invalid process value: {'id': 42} - object does not exist.",
        )

        self.assertEqual(Data.objects.count(), data_n)

    def test_post_no_perms(self):
        collection = Collection.objects.get(pk=1)
        collection.set_permission(Permission.VIEW, self.user2)

        data_count = Data.objects.count()
        resp = self._post(self.data, self.user2)
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(Data.objects.count(), data_count)

    def test_post_public_user(self):
        data_count = Data.objects.count()
        resp = self._post(self.data)
        # User has no permission to add Data object to the collection.
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(Data.objects.count(), data_count)

    def test_post_protected_fields(self):
        date_now = now()
        self.data["created"] = date_now - timedelta(days=360)
        self.data["modified"] = date_now - timedelta(days=180)
        self.data["started"] = date_now - timedelta(days=180)
        self.data["finished"] = date_now - timedelta(days=90)
        self.data["checksum"] = "fake"
        self.data["status"] = "DE"
        self.data["process_progress"] = 2
        self.data["process_rc"] = 18
        self.data["process_info"] = "Spam"
        self.data["process_warning"] = "More spam"
        self.data["process_error"] = "Even more spam"
        self.data["contributor_id"] = 2

        resp = self._post(self.data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)

        self.assertEqual(resp.data["started"], None)
        self.assertEqual(resp.data["finished"], None)
        self.assertEqual(
            resp.data["checksum"],
            "05bc76611c382a88817389019679f35cdb32ac65fe6662210805b588c30f71e6",
        )
        self.assertEqual(resp.data["status"], "RE")
        self.assertEqual(resp.data["process_progress"], 0)
        self.assertEqual(resp.data["process_rc"], None)
        self.assertEqual(resp.data["process_info"], [])
        self.assertEqual(resp.data["process_warning"], [])
        self.assertEqual(resp.data["process_error"], [])
        self.assertEqual(
            resp.data["contributor"],
            {
                "id": self.user1.pk,
                "username": self.user1.username,
                "first_name": self.user1.first_name,
                "last_name": self.user1.last_name,
            },
        )

    def test_post_contributor_numeric(self):
        response = ContributorSerializer(
            ContributorSerializer().to_internal_value(self.user1.pk)
        ).data

        self.assertEqual(
            response,
            {
                "id": self.user1.pk,
                "username": self.user1.username,
                "first_name": self.user1.first_name,
                "last_name": self.user1.last_name,
            },
        )

    def test_post_contributor_dict(self):
        response = ContributorSerializer(
            ContributorSerializer().to_internal_value({"id": self.user1.pk})
        ).data

        self.assertEqual(
            response,
            {
                "id": self.user1.pk,
                "username": self.user1.username,
                "first_name": self.user1.first_name,
                "last_name": self.user1.last_name,
            },
        )

    def test_post_contributor_dict_extra_data(self):
        response = ContributorSerializer(
            ContributorSerializer().to_internal_value(
                {"id": self.user1.pk, "username": "ignored", "first_name": "ignored"}
            )
        ).data

        self.assertEqual(
            response,
            {
                "id": self.user1.pk,
                "username": self.user1.username,
                "first_name": self.user1.first_name,
                "last_name": self.user1.last_name,
            },
        )

    def test_post_contributor_dict_invalid(self):
        with self.assertRaises(exceptions.ValidationError):
            ContributorSerializer().to_internal_value(
                {
                    "invalid-dictionary": True,
                }
            )

    def test_current_user_permissions(self):
        # Remove all permissions.
        self.data1.permission_group.permissions.all().delete()

        # Set permissions.
        self.data1.set_permission(Permission.VIEW, AnonymousUser())
        self.data1.set_permission(Permission.EDIT, self.user1)
        self.data1.set_permission(Permission.SHARE, self.user2)
        self.data1.set_permission(Permission.EDIT, self.admin)

        # Anonymous request, public view.
        expected_permissions = [{"type": "public", "permissions": ["view"]}]
        resp = self._get_detail(self.data1.pk)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertCountEqual(
            resp.data["current_user_permissions"], expected_permissions
        )

        # Anonymous request, no permissions.
        self.data1.set_permission(Permission.NONE, AnonymousUser())
        expected_permissions = [{}]
        resp = self._get_detail(self.data1.pk)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)
        self.data1.set_permission(Permission.VIEW, AnonymousUser())

        # User1 request.
        expected_permissions = [
            {"type": "public", "permissions": ["view"]},
            {
                "type": "user",
                "id": self.user1.pk,
                "name": self.user1.get_full_name() or self.user1.username,
                "username": self.user1.username,
                "permissions": ["view", "edit"],
            },
        ]
        resp = self._get_detail(self.data1.pk, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertCountEqual(
            resp.data["current_user_permissions"], expected_permissions
        )

        # User1 request, no permissions.
        self.data1.set_permission(Permission.NONE, self.user1)
        expected_permissions = [{"type": "public", "permissions": ["view"]}]
        resp = self._get_detail(self.data1.pk, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertCountEqual(
            resp.data["current_user_permissions"], expected_permissions
        )

        # User2 request.
        expected_permissions = [
            {"type": "public", "permissions": ["view"]},
            {
                "type": "user",
                "id": self.user2.pk,
                "name": self.user2.get_full_name() or self.user2.username,
                "username": self.user2.username,
                "permissions": ["view", "edit", "share"],
            },
        ]
        resp = self._get_detail(self.data1.pk, self.user2)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertCountEqual(
            resp.data["current_user_permissions"], expected_permissions
        )

        # Admin request: since admin is superuser all permissions are returned.
        expected_permissions = [
            {"type": "public", "permissions": ["view"]},
            {
                "type": "user",
                "id": self.admin.pk,
                "name": self.admin.get_full_name() or self.admin.username,
                "username": self.admin.username,
                "permissions": ["view", "edit", "share", "owner"],
            },
        ]
        resp = self._get_detail(self.data1.pk, self.admin)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertCountEqual(
            resp.data["current_user_permissions"], expected_permissions
        )

        # Admin request: since admin is superuser all permissions are returned
        # even if there are no permissions in the database.
        self.data1.permission_group.permissions.all().delete()
        expected_permissions = [
            {
                "type": "user",
                "id": self.admin.pk,
                "name": self.admin.get_full_name() or self.admin.username,
                "username": self.admin.username,
                "permissions": ["view", "edit", "share", "owner"],
            },
        ]
        resp = self._get_detail(self.data1.pk, self.admin)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertCountEqual(
            resp.data["current_user_permissions"], expected_permissions
        )

        # Check listing.
        self.data1.permission_group.permissions.all().delete()
        self.data1.set_permission(Permission.VIEW, self.user1)
        data2 = Data.objects.get(pk=2)
        data2.permission_group.permissions.all().delete()
        data2.set_permission(Permission.EDIT, self.user1)
        data2.set_permission(Permission.VIEW, AnonymousUser())
        expected_permissions = {
            self.data1.pk: [
                {
                    "type": "user",
                    "id": self.user1.pk,
                    "name": self.user1.get_full_name() or self.user1.username,
                    "username": self.user1.username,
                    "permissions": ["view"],
                }
            ],
            data2.pk: [
                {
                    "type": "user",
                    "id": self.user1.pk,
                    "name": self.user1.get_full_name() or self.user1.username,
                    "username": self.user1.username,
                    "permissions": ["view", "edit"],
                },
                {"type": "public", "permissions": ["view"]},
            ],
        }

        resp = self._get_list(self.user1)
        self.assertEqual(len(resp.data), 2)
        for i in range(2):
            self.assertCountEqual(
                resp.data[i]["current_user_permissions"],
                expected_permissions[resp.data[i]["id"]],
            )

        # Admin should have all permissions.
        resp = self._get_list(self.admin)
        expected_permissions = {
            self.data1.pk: [
                {
                    "type": "user",
                    "id": self.admin.pk,
                    "name": self.admin.get_full_name() or self.admin.username,
                    "username": self.admin.username,
                    "permissions": ["view", "edit", "share", "owner"],
                }
            ],
            data2.pk: [
                {
                    "type": "user",
                    "id": self.admin.pk,
                    "name": self.admin.get_full_name() or self.admin.username,
                    "username": self.admin.username,
                    "permissions": ["view", "edit", "share", "owner"],
                },
                {"type": "public", "permissions": ["view"]},
            ],
        }
        self.assertEqual(len(resp.data), 2)
        for i in range(2):
            self.assertCountEqual(
                resp.data[i]["current_user_permissions"],
                expected_permissions[resp.data[i]["id"]],
            )

    def test_get_detail(self):
        # public user w/ perms
        resp = self._get_detail(1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(resp.data["id"], 1)

        # user w/ permissions
        resp = self._get_detail(1, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertKeys(
            resp.data,
            [
                "slug",
                "name",
                "created",
                "modified",
                "contributor",
                "started",
                "finished",
                "checksum",
                "status",
                "process",
                "process_progress",
                "process_rc",
                "process_info",
                "process_warning",
                "process_error",
                "input",
                "output",
                "descriptor_schema",
                "descriptor",
                "id",
                "size",
                "scheduled",
                "current_user_permissions",
                "descriptor_dirty",
                "tags",
                "process_memory",
                "process_cores",
                "collection",
                "entity",
                "duplicated",
                "process_resources",
            ],
        )

        # user w/ public permissions
        resp = self._get_detail(1, self.user2)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertKeys(
            resp.data,
            [
                "slug",
                "name",
                "created",
                "modified",
                "contributor",
                "started",
                "finished",
                "checksum",
                "status",
                "process",
                "process_progress",
                "process_rc",
                "process_info",
                "process_warning",
                "process_error",
                "input",
                "output",
                "descriptor_schema",
                "descriptor",
                "id",
                "size",
                "scheduled",
                "current_user_permissions",
                "descriptor_dirty",
                "tags",
                "process_memory",
                "process_cores",
                "collection",
                "entity",
                "duplicated",
                "process_resources",
            ],
        )

    def test_get_detail_no_perms(self):
        # public user w/o permissions
        resp = self._get_detail(2)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(resp.data["detail"], MESSAGES["NOT_FOUND"])

        # user w/o permissions
        resp = self._get_detail(2, self.user2)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(resp.data["detail"], MESSAGES["NOT_FOUND"])

    def test_patch(self):
        data = {"name": "New data"}
        resp = self._patch(1, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.name, "New data")

    def test_patch_no_perms(self):
        data = {"name": "New data"}
        resp = self._patch(2, data, self.user2)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)
        d = Data.objects.get(pk=2)
        self.assertEqual(d.name, "Test data 2")

    def test_patch_public_user(self):
        data = {"name": "New data"}
        resp = self._patch(2, data)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)
        d = Data.objects.get(pk=2)
        self.assertEqual(d.name, "Test data 2")

    def test_patch_protected(self):
        date_now = now()

        # `created`
        resp = self._patch(1, {"created": date_now}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.created.isoformat(), self.data1.created.isoformat())
        self.assertEqual(d.modified.isoformat(), self.data1.modified.isoformat())

        # `modified`
        resp = self._patch(1, {"modified": date_now - timedelta(days=180)}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.modified.isoformat(), self.data1.modified.isoformat())

        # `started`
        resp = self._patch(1, {"started": date_now}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.started.isoformat(), self.data1.started.isoformat())
        self.assertEqual(d.modified.isoformat(), self.data1.modified.isoformat())

        # `finished`
        resp = self._patch(1, {"finished": date_now}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.finished.isoformat(), self.data1.finished.isoformat())
        self.assertEqual(d.modified.isoformat(), self.data1.modified.isoformat())

        # `checksum`
        resp = self._patch(1, {"checksum": "fake"}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(
            d.checksum,
            "05bc76611c382a88817389019679f35cdb32ac65fe6662210805b588c30f71e6",
        )
        self.assertEqual(d.modified.isoformat(), self.data1.modified.isoformat())

        # `status`
        resp = self._patch(1, {"status": "DE"}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.status, "OK")
        self.assertEqual(d.modified.isoformat(), self.data1.modified.isoformat())

        # `process_progress`
        resp = self._patch(1, {"process_progress": 2}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.process_progress, 0)
        self.assertEqual(d.modified.isoformat(), self.data1.modified.isoformat())

        # `process_rc`
        resp = self._patch(1, {"process_rc": 18}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.process_rc, None)
        self.assertEqual(d.modified.isoformat(), self.data1.modified.isoformat())

        # `process_info`
        resp = self._patch(1, {"process_info": "Spam"}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.process_info, [])
        self.assertEqual(d.modified.isoformat(), self.data1.modified.isoformat())

        # `process_warning`
        resp = self._patch(1, {"process_warning": "More spam"}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.process_warning, [])
        self.assertEqual(d.modified.isoformat(), self.data1.modified.isoformat())

        # `process_error`
        resp = self._patch(1, {"process_error": "Even more spam"}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.process_error, [])
        self.assertEqual(d.modified.isoformat(), self.data1.modified.isoformat())

        # `contributor`
        resp = self._patch(1, {"contributor": 2}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.contributor_id, 1)
        self.assertEqual(d.modified.isoformat(), self.data1.modified.isoformat())

        # `process`
        resp = self._patch(1, {"process": 2}, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        d = Data.objects.get(pk=1)
        self.assertEqual(d.process_id, 1)
        self.assertEqual(d.modified.isoformat(), self.data1.modified.isoformat())


class DataTestCaseDelete(
    DataTestCaseCommonMixin, TransactionResolweAPITestCase, DjangoTransactionTestCase
):
    def _pre_setup(self, *args, **kwargs):
        """Delete all previously created users.

        The public user is created in the migrations and recreated during test case,
        but with a different id which breaks the loading of the fixtures.
        """
        get_user_model().objects.all().delete()
        super()._pre_setup(*args, **kwargs)

    def test_delete(self):
        response = self._delete(1, self.user1)
        BackgroundTask.objects.get(pk=response.data["id"]).wait()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertFalse(Data.objects.filter(pk=1).exists())

    def test_bulk_delete(self):
        request_path = reverse("resolwe-api:data-bulk-delete")
        for data in Data.objects.all():
            data.set_permission(Permission.NONE, get_anonymous_user(False))
            data.set_permission(Permission.NONE, self.user1)

        # Anonymous request, no permission.
        response = self.client.post(
            request_path, data={"ids": [self.data1.pk]}, format="json"
        )
        self.assertContains(
            response,
            f"No permission to delete objects with ids {set([self.data1.pk])}",
            status_code=response.status_code,
        )
        self.assertTrue(Data.objects.filter(pk=1).exists())
        self.assertTrue(Data.objects.filter(pk=2).exists())

        # Anonymous request, view permission.
        self.data1.set_permission(Permission.VIEW, get_anonymous_user(False))
        response = self.client.post(
            request_path, data={"ids": [self.data1.pk]}, format="json"
        )
        self.assertContains(
            response,
            f"No permission to delete objects with ids {self.data1.pk}.",
            status_code=status.HTTP_403_FORBIDDEN,
        )
        self.assertTrue(Data.objects.filter(pk=1).exists())
        self.assertTrue(Data.objects.filter(pk=2).exists())
        self.data1.set_permission(Permission.NONE, get_anonymous_user(False))

        # All requests are authenticated as user1 from now on.
        self.client.force_authenticate(self.user1)

        # No permission.
        self.data1.set_permission(Permission.NONE, self.user1)
        self.data2.set_permission(Permission.NONE, self.user1)
        response = self.client.post(
            request_path, data={"ids": [self.data1.pk, self.data2.pk]}, format="json"
        )
        self.assertContains(
            response,
            "No permission to delete objects with ids",
            status_code=response.status_code,
        )
        self.assertTrue(Data.objects.filter(pk=1).exists())
        self.assertTrue(Data.objects.filter(pk=2).exists())

        # View permission.
        self.data1.set_permission(Permission.VIEW, self.user1)
        self.data2.set_permission(Permission.VIEW, self.user1)
        response = self.client.post(
            request_path, data={"ids": [self.data1.pk, self.data2.pk]}, format="json"
        )
        self.assertContains(
            response,
            f"No permission to delete objects with ids {self.data1.pk}, {self.data2.pk}.",
            status_code=status.HTTP_403_FORBIDDEN,
        )
        self.assertTrue(Data.objects.filter(pk=1).exists())
        self.assertTrue(Data.objects.filter(pk=2).exists())

        # Partial permission.
        self.data1.set_permission(Permission.EDIT, self.user1)
        self.data2.set_permission(Permission.NONE, self.user1)
        response = self.client.post(request_path, data={"ids": [1, 2]}, format="json")
        BackgroundTask.objects.get(pk=response.data["id"]).wait()
        self.assertFalse(Data.objects.filter(pk=1).exists())
        self.assertTrue(Data.objects.filter(pk=2).exists())

        # All permissions.
        self.data2.set_permission(Permission.EDIT, self.user1)
        response = self.client.post(request_path, data={"ids": [1, 2]}, format="json")
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        BackgroundTask.objects.get(pk=response.data["id"]).wait()
        self.assertFalse(Data.objects.filter(pk=1).exists())
        self.assertFalse(Data.objects.filter(pk=2).exists())

    def test_delete_no_perms(self):
        response = self._delete(2, self.user2)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        query = Data.objects.filter(pk=2).exists()
        self.assertTrue(query)

    def test_delete_public_user(self):
        response = self._delete(2)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        query = Data.objects.filter(pk=2).exists()
        self.assertTrue(query)



"""Application configuration."""

from django.apps import AppConfig


class BaseConfig(AppConfig):
    """Application configuration."""

    name = "resolwe.observers"

    def ready(self):
        """Application initialization."""
        # Register signals handlers.
        from . import signals  # noqa: F401



"""Processing container startup script."""

import array
import asyncio
import concurrent
import json
import logging
import os
import select
import shlex
import shutil
import signal
import socket
import sys
import uuid
from contextlib import closing, suppress
from distutils.util import strtobool
from functools import partial
from pathlib import Path
from time import time
from typing import Dict, List, Optional, Sequence

try:
    import constants
    from socket_utils import (
        BaseCommunicator,
        BaseProtocol,
        Message,
        PeerIdentity,
        Response,
        SocketCommunicator,
    )

except ImportError:
    from resolwe.flow.executors import constants
    from resolwe.flow.executors.socket_utils import (
        BaseCommunicator,
        BaseProtocol,
        Message,
        PeerIdentity,
        Response,
        SocketCommunicator,
    )

# The directory where local processing is done.
TMP_DIR = Path(os.environ.get("TMP_DIR", ".tmp"))
DATA_ID = int(os.getenv("DATA_ID", "-1"))

# Basic constants.
WORKING_DIRECTORY = Path.cwd().resolve()
STDOUT_LOG_PATH = WORKING_DIRECTORY / "stdout.txt"
JSON_LOG_PATH = WORKING_DIRECTORY / "jsonout.txt"

# Update log files every 30 seconds.
UPDATE_LOG_FILES_TIMEOUT = 30

# Keep data settings.
KEEP_DATA = bool(strtobool(os.environ.get("FLOW_MANAGER_KEEP_DATA", "False")))

# Read configuration from environmental variables.
COMMUNICATION_SOCKET = (
    constants.SOCKETS_VOLUME / constants.COMMUNICATION_PROCESSING_SOCKET
)
SCRIPT_SOCKET = constants.SOCKETS_VOLUME / constants.SCRIPT_SOCKET
UPLOAD_FILE_SOCKET = constants.SOCKETS_VOLUME / constants.UPLOAD_FILE_SOCKET

# How many file descriptors to sent over socket in a single message.
DESCRIPTOR_CHUNK_SIZE = int(os.environ.get("DESCRIPTOR_CHUNK_SIZE", 100))

# Configure container log level.
LOG_LEVEL = int(os.environ.get("LOG_LEVEL", logging.DEBUG))

# Max characters to log in a single message: the processing script messages can
# be hundreds lines long.
LOG_MAX_LENGTH = int(os.environ.get("LOG_MAX_LENGTH", 200))

# How long to wait for the response from the communication container.
# Do not set too low or script may fail to start. The default is 20 minutes.
SCRIPT_START_TIMEOUT = int(os.environ.get("SCRIPT_START_TIMEOUT", 1200))

logging.basicConfig(
    stream=sys.stdout,
    level=LOG_LEVEL,
    format=f"%(asctime)s - %(name)s - %(levelname)s - %(message).{LOG_MAX_LENGTH}s",
)
logger = logging.getLogger(__name__)


assert sys.version_info[0] == 3, "Only Python3 is supported"
assert sys.version_info[1] >= 6, "Only Python >= 3.6 is supported"


async def stop_task(task: Optional[asyncio.Task], timeout: int = 1):
    """Cancel the given task.

    Also wait for task to stop for up to timeout seconds.
    """
    if task is not None and not task.done():
        task.cancel()
        with suppress(asyncio.CancelledError):
            await asyncio.wait_for(task, timeout=timeout)


async def log_exception(
    message: str, communicator: BaseCommunicator, timeout: int = 60
):
    """Log the exception via logger and send it to the communicator.

    Log it using logger and send error message to the communication
    container. Make sure no exception is ever raised and that the method
    completes in timeout seconds.
    """
    with suppress(Exception):
        logger.exception(message)
    with suppress(Exception):
        log_coroutine = communicator.process_log({"error": [message]})

        await asyncio.wait_for(log_coroutine, timeout)


async def heartbeat_task(communicator: BaseCommunicator, heartbeat_interval: int = 60):
    """Periodically send heartbeats.

    This task will run indefinitely and has to be cancelled to stop.

    :attr heartbeat_interval: seconds between successive heartbeat messages.
    """
    while True:
        await communicator.send_heartbeat()
        await asyncio.sleep(heartbeat_interval)


async def log_error(message: str, communicator: BaseCommunicator, timeout: int = 60):
    """Log the error via logger and send it to the communicator.

    Log it using logger and send error message to the communication
    container. Make sure no exception is ever raised and that the method
    completes in timeout seconds.
    """
    with suppress(Exception):
        logger.error(message)
    with suppress(Exception):
        log_coroutine = communicator.process_log({"error": [message]})

        await asyncio.wait_for(log_coroutine, timeout)


async def connect_to_communication_container(path: str) -> SocketCommunicator:
    """Wait for the communication container to start.

    Assumption is that the communication container is accepting connections
    and will send a single PING when connected.

    The coroutine will only exit when the connection is successfully
    established it is the responsibility of the caller to terminate the
    coroutine when the connection can not be established for some time.
    """
    while True:
        with suppress(Exception):
            reader, writer = await asyncio.open_unix_connection(path)
            line = (await reader.readline()).decode("utf-8")
            assert line.strip() == "PING", "Expected 'PING', got '{}'.".format(line)
            return SocketCommunicator(
                reader, writer, "processing<->communication", logger
            )
        await asyncio.sleep(1)


async def start_script(script: str) -> asyncio.subprocess.Process:
    """Start processing the script.

    :raises Exception: when script could not be started.
    """
    # Start the bash subprocess and write script to its standard input.
    logger.debug("Executing script: '%s'.", script)
    bash_command = "/bin/bash --login" + os.linesep
    proc = await asyncio.subprocess.create_subprocess_exec(
        *shlex.split(bash_command),
        limit=4 * (2**20),  # 4MB buffer size for line buffering
        stdin=asyncio.subprocess.PIPE,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.STDOUT,
    )
    # Assert that stdin and stdout streams are connected to the proc.
    error_message = "The stream '{}' could not connect to the script."
    assert proc.stdin is not None, error_message.format("stdin")
    assert proc.stdout is not None, error_message.format("stdout")
    proc.stdin.write(script.encode("utf-8"))
    proc.stdin.close()
    logger.debug("Code sent to stdin of the script process.")
    return proc


async def communicate_with_process(
    proc: asyncio.subprocess.Process, manager: "ProcessingManager"
) -> int:
    """Read the given process standard output and write it to the log file.

    :return: the process's return code.
    """
    try:
        assert proc.stdout is not None
        while not proc.stdout.at_eof():
            # Here we tap into the internal structures of the StreamReader
            # in order to read as much data as we can in one chunk.
            await proc.stdout._wait_for_data("execute_script")  # type: ignore
            # Send received data to upload manager.
            manager.upload_handler.update_stdout(proc.stdout._buffer)
            proc.stdout._buffer.clear()
            # Break the loop if EOF is reached.
            if proc.stdout.at_eof():
                logger.debug("Processing script stdout eof.")
                break
        # Wait for the script to terminate.
        await proc.wait()

    except Exception:
        logger.exception("Exception running the script process.")
        # Terminate process if it is still running.
        with suppress(Exception):
            if proc.returncode is None:
                proc.terminate()
                with suppress(asyncio.TimeoutError):
                    await asyncio.wait_for(proc.wait(), timeout=60)
            if proc.returncode is None:
                proc.kill()
                with suppress(asyncio.TimeoutError):
                    await asyncio.wait_for(proc.wait(), timeout=60)
    finally:
        # Return the return code (if available) or general error code 1.
        return_code = proc.returncode if proc.returncode is not None else 1
        logger.debug("Script finished with the return code '%d'.", return_code)
        manager.terminate()
        return return_code


async def initialize_connections(
    loop: asyncio.AbstractEventLoop,
) -> "ProcessingManager":
    """Start the processing manager.

    This method does the following:
    - connects to the communication container,
    - connects to the upload socket (unix socket) and
    - starts the socket server for the communication with the processing
        script.

    :raises ConnectionError: when connection can not be established.
    :return: the processing manager instance.
    """
    # Connect to the communication container and start processing script
    # server first.
    logger.debug(
        "Connecting to the communication container via socket '%s'.",
        COMMUNICATION_SOCKET,
    )
    start = time()
    communicator_future: asyncio.Task = asyncio.ensure_future(
        connect_to_communication_container(str(COMMUNICATION_SOCKET))
    )
    try:
        socket_communicator = await asyncio.wait_for(
            communicator_future, timeout=constants.CONTAINER_TIMEOUT
        )
    except (asyncio.CancelledError, asyncio.TimeoutError):
        raise ConnectionError("Timeout connecting to communication container.")
    logger.debug(
        "Connected to the communication container in '%.2f' seconds.",
        time() - start,
    )

    # Connect to the upload socket and start the periodic upload of log files.
    upload_socket = socket.socket(family=socket.AF_UNIX)
    upload_socket.connect(str(UPLOAD_FILE_SOCKET))
    upload_handler = UploadHandler(upload_socket, loop)
    logger.debug("Connected to the upload socket '%s'.", UPLOAD_FILE_SOCKET)
    upload_handler.start_log_upload()

    # Start the server to handle processing script connections.
    logger.debug(
        "Starting the processing script server on the socket '%s'.", SCRIPT_SOCKET
    )
    script_server_future: asyncio.Task = asyncio.ensure_future(
        asyncio.start_unix_server(
            partial(
                handle_processing_script_connection,
                socket_communicator,
                upload_handler,
            ),
            str(SCRIPT_SOCKET),
        )
    )
    try:
        script_server = await asyncio.wait_for(script_server_future, timeout=60)
    except asyncio.CancelledError:
        raise ConnectionError("Timeout starting processing script server.")

    return ProcessingManager(socket_communicator, upload_handler, script_server, loop)


async def handle_processing_script_connection(
    communication_communicator: SocketCommunicator,
    upload_handler: "UploadHandler",
    reader: asyncio.StreamReader,
    writer: asyncio.StreamWriter,
):
    """Handle incoming connection from the processing script.

    Python process opens a single connection while Resolwe runtime utils opens
    a new connection for every request.
    """
    try:
        logger.info("Processing script connected to the socket.")
        communicator = SocketCommunicator(reader, writer, "script", logger)
        protocol_handler = ScriptProtocol(
            communication_communicator, communicator, upload_handler
        )
        await protocol_handler.communicate()
    except (asyncio.IncompleteReadError, GeneratorExit):
        logger.info("Processing script closed connection.")
    except Exception:
        logger.exception("Unknown exception while handling processing script message.")
        raise


class CommunicationProtocol(BaseProtocol):
    """Handles communication with the communication container."""

    def __init__(self, manager: "ProcessingManager"):
        """Initialize variables."""
        super().__init__(manager.communicator, logger)
        self.processing_manager = manager
        self.script_task: Optional[asyncio.Task] = None
        self._script_starting = asyncio.Lock()
        self.script_started = asyncio.Event()

    async def handle_process_script(
        self, message: Message, identity: PeerIdentity
    ) -> Response[int]:
        """Handle the process_script message."""
        # This message can arrive multiple times so make sure the script is
        # started only once.
        if self.script_started.is_set():
            return message.respond_ok("Already processing")
        self.script_started.set()
        if self.script_task is None:
            process = await start_script(message.message_data)
            self.script_task = asyncio.ensure_future(
                communicate_with_process(process, self.processing_manager)
            )
        return message.respond_ok("Script started.")

    async def terminate_script(self):
        """Terminate the script."""
        await stop_task(self.script_task)

    async def post_terminate(self, message: Message, identity: PeerIdentity):
        """Stop the processing script and stop the manager."""
        logger.debug("Stopping all the tasks in the post_terminate handler.")
        await stop_task(self.script_task)
        self.processing_manager.terminate()

    async def handle_terminate(
        self, message: Message, identity: PeerIdentity
    ) -> Response[str]:
        """Handle the terminate command."""
        logger.debug("Received terminate command from the communication container.")
        return message.respond_ok("Terminating")


class ProcessingManager:
    """Manager class.

    It starts network servers and executes the received script.
    """

    def __init__(
        self,
        communicator: BaseCommunicator,
        upload_handler: "UploadHandler",
        script_server,
        loop: asyncio.AbstractEventLoop,
    ):
        """Initialize variables, start the communicator and heartbeat task."""
        self.upload_handler = upload_handler
        self.communicator = communicator
        self.loop = loop
        self.protocol_handler = CommunicationProtocol(self)
        self.script_server = script_server

        self.communicator_task = asyncio.ensure_future(
            self.protocol_handler.communicate()
        )
        self.heartbeat_task = asyncio.ensure_future(heartbeat_task(communicator))

        # TODO: drop this when we stop supporting Python 3.6.
        # This class has been implicitly getting the current running loop since 3.7.
        lock_arguments: Dict[str, asyncio.AbstractEventLoop] = {}
        if sys.version_info < (3, 7):
            lock_arguments = {"loop": loop}
        self._terminating = asyncio.Event(**lock_arguments)

    def terminate(self):
        """Terminate the processing container."""
        self._terminating.set()

    async def _stop_all_tasks(self, stop_timeout=60):
        """Stop all running tasks."""
        # Stop the processing script.
        logger.debug("Stopping all the tasks in the processing manager.")
        logger.debug("Stopping the processing script task.")
        await stop_task(self.protocol_handler.script_task)

        # Stop the script server.
        logger.debug("Stopping the script server.")
        with suppress(Exception):
            self.script_server.close()
            await asyncio.wait_for(
                self.script_server.wait_closed(), timeout=stop_timeout
            )

        # Stop the periodic log upload task.
        logger.debug("Stopping periodic log upload task.")
        await self.upload_handler.stop_log_upload()

        # Upload latest log files (in case there was a change). Notify the
        # listener on failure.
        logger.debug("Uploading latest log files.")
        try:
            # Give it up to one minute to upload the latest log files.
            latest_log_upload = asyncio.ensure_future(
                self.upload_handler.upload_log_files()
            )
            await asyncio.wait_for(latest_log_upload, timeout=60)
        except Exception:
            await log_exception("Unable to upload the latest logs.", self.communicator)

        # If data should be kept it must be uploaded to the data dir.
        if KEEP_DATA:
            logger.debug("Flag KEEP_DATA is set, uploading files.")
            try:
                # The KEEP_DATA setting is used only for debugging purposes so the
                # task should not be used to transfer very large files.
                # We wait for up to 10 minutes for it to complete.
                send_coroutine = self.upload_handler.send_file_descriptors(
                    [str(file) for file in Path("./").rglob("*") if file.is_file()]
                )
                await asyncio.wait_for(send_coroutine, timeout=600)
            except Exception:
                await log_exception(
                    "Unable to upload the data (KEEP_DATA flag is set).",
                    self.communicator,
                )

        # Send the finish command to the communication container.
        logger.debug("Send the finish command to the communication container.")
        await self.communicator.finish(self.return_code)

        # Stop the heartbeat task.
        logger.debug("Stopping the heartbeat task.")
        await stop_task(self.heartbeat_task)

        # Finally, stop the communicator task.
        with suppress(Exception):
            logger.debug("Stopping communication with the communication container.")
            self.protocol_handler.stop_communicate()
            await asyncio.wait_for(self.communicator_task, timeout=stop_timeout)
        logger.debug("All tasks stopped.")

    @property
    def return_code(self):
        """Get the return code.

        :return: return the script task return code. When the script task is
            still running return 1.
        """
        task = self.protocol_handler.script_task
        if task is None or task.result() is None:
            return 1
        return task.result()

    async def run(self) -> int:
        """Start the main method.

        :raises RuntimeError: in case of failure.
        """
        try:
            loop = asyncio.get_event_loop()
            await asyncio.wait(
                [
                    loop.create_task(self.protocol_handler.script_started.wait()),
                    loop.create_task(self._terminating.wait()),
                ],
                timeout=SCRIPT_START_TIMEOUT,
                return_when=asyncio.FIRST_COMPLETED,
            )
            if self._terminating.is_set():
                logger.debug("Terminating message received.")
            elif not self.protocol_handler.script_started.is_set():
                await log_error(
                    f"The script failed to start in {SCRIPT_START_TIMEOUT} seconds.",
                    self.communicator,
                )
            else:
                monitored_tasks = (
                    loop.create_task(self.script_server.wait_closed()),
                    self.upload_handler.upload_task,
                    self.communicator_task,
                    loop.create_task(self._terminating.wait()),
                )
                # Stop processing when any of the following tasks complete.
                await asyncio.wait(monitored_tasks, return_when=asyncio.FIRST_COMPLETED)
        except Exception:
            await log_exception(
                "Unexpected exception in the processing container.", self.communicator
            )
        finally:
            # Send the return code to the communicate container.
            await self._stop_all_tasks()
            return self.return_code


class UploadHandler:
    """Handle upload of files via communication container."""

    def __init__(
        self, upload_socket: socket.socket, loop: asyncio.AbstractEventLoop
    ) -> None:
        """Initialize local variables."""
        self._log_buffers = {"JSON": bytearray(), "STDOUT": bytearray()}
        self._log_files = {"JSON": JSON_LOG_PATH, "STDOUT": STDOUT_LOG_PATH}

        self.upload_socket = upload_socket
        self.exported_files_mapper: Dict[str, str] = dict()
        self.upload_task: Optional[asyncio.Task] = None
        lock_arguments: Dict[str, asyncio.AbstractEventLoop] = dict()
        if sys.version_info < (3, 7):
            lock_arguments = {"loop": loop}
        self._uploading_log_files_lock = asyncio.Lock(**lock_arguments)
        self._send_file_descriptors_lock = asyncio.Lock(**lock_arguments)
        self.loop = loop

    async def export_files(self, file_names: Sequence[str]):
        """Handle export files."""
        unique_names = []
        for file_name in file_names:
            unique_name = "export_" + uuid.uuid4().hex
            shutil.move(file_name, unique_name)
            unique_names.append(unique_name)
        presigned_urls = await self.send_file_descriptors(
            unique_names, need_presigned_urls=True, storage_name="upload"
        )
        for file_name, presigned_url in zip(file_names, presigned_urls):
            self.exported_files_mapper[file_name] = presigned_url

    def update_stdout(self, data: bytes):
        """Write data bytes to stdout."""
        self._log_buffers["STDOUT"] += data

    def update_jsonout(self, data: bytes):
        """Write data bytes to jsonout."""
        self._log_buffers["JSON"] += data

    def start_log_upload(self) -> asyncio.Task:
        """Start the timer that periodically uploads the log files.

        On multiple calls already created task is returned.
        """
        if self.upload_task is None or self.upload_task.done():
            self.upload_task = asyncio.ensure_future(self._update_logs_timer())
        return self.upload_task

    async def stop_log_upload(self):
        """Stop the timer that periodically uploads the log files."""
        if self.upload_task is not None:
            # Only cancel the task when uploading is not in progress.
            async with self._uploading_log_files_lock:
                await stop_task(self.upload_task)

    async def _update_logs_timer(self, stop_after: int = 3):
        """Update log files every UPDATE_LOG_FILES_TIMEOUT secods.

        :attr stop_after: the integer indication how many exceptions must occur
            during log upload to terminate the timer task.

        :raises Exception: when stop_after exceptions occur during transfer the
            last exception is re-raised.
        """
        exception_count = 0
        while True:
            try:
                await asyncio.sleep(UPDATE_LOG_FILES_TIMEOUT)
                logger.debug("Uploading log timer triggered.")
                await self.upload_log_files()
            except asyncio.CancelledError:
                logger.debug("Update log files timer canceled.")
                break
            except Exception:
                logger.exception("Error uploading log files.")
                exception_count += 1
                if exception_count >= stop_after:
                    raise

    async def upload_log_files(self):
        """Upload log files if needed.

        :raises BrokenPipeError: when upload socket is not available.
        """
        to_upload = []
        for log_type in ["JSON", "STDOUT"]:
            if self._log_buffers[log_type]:
                to_upload.append(
                    self._log_files[log_type].relative_to(WORKING_DIRECTORY)
                )
                with self._log_files[log_type].open("ba") as log_file:
                    log_file.write(self._log_buffers[log_type])
                    self._log_buffers[log_type].clear()

        if to_upload:
            logger.debug("Uploading log files: %s.", to_upload)
            async with self._uploading_log_files_lock:
                await self.send_file_descriptors(to_upload)

    async def send_file_descriptors(
        self,
        filenames: Sequence[str],
        need_presigned_urls: bool = False,
        storage_name: str = "data",
    ):
        """Send file descriptors over UNIX sockets.

        Code is based on the sample in manual
        https://docs.python.org/3/library/socket.html#socket.socket.sendmsg .

        :returns: the list of presigned urls for filenames (in the same order)
            if need_presigned_urls is set to True. Otherwise empty list is
            returned.

        :raises RuntimeError: on failure.
        """

        def receive(length: int) -> bytes:
            """Receive the length bytes from the upload socket."""
            received = 0
            message = b""
            while received != length:
                select.select([self.upload_socket], [], [])
                payload = self.upload_socket.recv(length - received)
                if payload == b"":
                    raise RuntimeError("Error reading data from the upload socket.")
                message += payload
                received = len(message)
            return message

        def send(payload: bytes):
            """Send the payload over the upload socket."""
            bytes_sent = 0
            to_send = len(payload)
            while bytes_sent < to_send:
                select.select([], [self.upload_socket], [])
                chunk_size = self.upload_socket.send(payload)
                bytes_sent += chunk_size
                payload = payload[chunk_size:]
                if chunk_size == 0:
                    raise RuntimeError("Error sending data to the upload socket.")

        def send_chunk(processing_filenames: Sequence[str]) -> List[str]:
            """Send chunk of filenames to uploader.

            :raises RuntimeError: on failure.
            :returns: the list of presigned urls (if requested) or empty list.
            """
            logger.debug("Sending file descriptors for files: %s", processing_filenames)
            message = [storage_name, processing_filenames, need_presigned_urls]
            payload = json.dumps(message).encode()

            # File handlers must be created before file descriptors otherwise
            # garbage collector will kick in and close the handlers. Handlers
            # will be closed when function completes.
            file_handlers = [open(filename, "rb") for filename in processing_filenames]
            file_descriptors = [file_handler.fileno() for file_handler in file_handlers]
            send(len(payload).to_bytes(8, byteorder="big"))
            select.select([], [self.upload_socket], [])
            self.upload_socket.sendmsg(
                [payload],
                [
                    (
                        socket.SOL_SOCKET,
                        socket.SCM_RIGHTS,
                        array.array("i", file_descriptors),
                    )
                ],
            )
            response_length = int.from_bytes(receive(8), byteorder="big")
            response = json.loads(receive(response_length).decode())
            if not response["success"]:
                raise RuntimeError(
                    "Communication container response indicates error sending "
                    "files {}.".format(processing_filenames)
                )
            return response["presigned_urls"]

        result = []
        self.upload_socket.setblocking(False)
        async with self._send_file_descriptors_lock:
            with concurrent.futures.ThreadPoolExecutor(max_workers=1) as pool:
                for i in range(0, len(filenames), DESCRIPTOR_CHUNK_SIZE):
                    processing_filenames = [
                        str(file_) for file_ in filenames[i : i + DESCRIPTOR_CHUNK_SIZE]
                    ]
                    response = await self.loop.run_in_executor(
                        pool, send_chunk, processing_filenames
                    )
                    if need_presigned_urls:
                        result += response
        return result

    def close(self):
        """Close the upload socket."""
        self.upload_socket.close()


class ScriptProtocol(BaseProtocol):
    """Process the processing script messages."""

    def __init__(
        self,
        communication_communicator: SocketCommunicator,
        script_communicator: SocketCommunicator,
        upload_handler: UploadHandler,
    ):
        """Initialize."""
        super().__init__(script_communicator, logger)
        self.communication_communicator = communication_communicator
        self.upload_handler = upload_handler

    async def process_command(
        self, peer_identity: PeerIdentity, received_message: Message
    ):
        """Log the received message and call the parents handler."""
        self.upload_handler.update_jsonout(
            json.dumps(received_message.to_dict()).encode()
        )
        return await super().process_command(peer_identity, received_message)

    async def handle_export_files(
        self, message: Message, identity: PeerIdentity
    ) -> Response[str]:
        """Export files by sending them to the communication container."""
        await self.upload_handler.export_files(message.message_data)
        return message.respond_ok("")

    async def handle_upload_files(
        self, message: Message, identity: PeerIdentity
    ) -> Response[str]:
        """Upload files by sending them to the communication container."""
        filenames = message.message_data
        await self.upload_handler.send_file_descriptors(filenames)
        return message.respond_ok("")

    async def handle_run(
        self, message: Message, identity: PeerIdentity
    ) -> Response[str]:
        """Handle the run command."""
        command_data = {
            "data": message.message_data,
            "export_files_mapper": self.upload_handler.exported_files_mapper,
        }
        command = Message.command(message.command_name, command_data)
        return await self.communication_communicator.send_command(command)

    async def default_command_handler(
        self, message: Message, identity: PeerIdentity
    ) -> Response:
        """By default proxy all commands to the communication container."""
        return await self.communication_communicator.send_command(message)


def sig_term_handler(processing_manager: ProcessingManager):
    """Gracefully terminate the running process."""
    logger.debug("SIG_INT received, shutting down.")
    processing_manager.terminate()


async def init(loop: asyncio.AbstractEventLoop) -> ProcessingManager:
    """Create necessary objects."""
    # Create temporary directory inside working directory.
    temporary_directory = WORKING_DIRECTORY / TMP_DIR
    temporary_directory.mkdir(exist_ok=True)

    # Create (empty) log files.
    STDOUT_LOG_PATH.touch()
    JSON_LOG_PATH.touch()

    # Create the processing manager.
    processing_manager = await initialize_connections(loop)

    # Addd signal handler for SIGINT and SIGTERM.
    loop.add_signal_handler(
        signal.SIGINT, partial(sig_term_handler, processing_manager)
    )
    loop.add_signal_handler(
        signal.SIGTERM, partial(sig_term_handler, processing_manager)
    )

    return processing_manager


if __name__ == "__main__":
    # Replace the code bellow with asyncio.run when we no longer have to
    # support python 3.6.
    with closing(asyncio.get_event_loop()) as loop:
        processing_manager = loop.run_until_complete(init(loop))
        return_code = loop.run_until_complete(processing_manager.run())
    sys.exit(return_code)



"""Resolwe annotations model."""

import abc
import datetime
import re
from collections import defaultdict
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    List,
    Mapping,
    MutableSequence,
    Optional,
    Sequence,
    Type,
)

from django.core.exceptions import ValidationError
from django.db import models
from versionfield import VersionField

from resolwe.flow.models.base import (
    MAX_SLUG_LENGTH,
    BaseManager,
    BaseModel,
    ResolweSlugField,
)
from resolwe.permissions.models import (
    PermissionInterface,
    PermissionObject,
    PermissionQuerySet,
)

if TYPE_CHECKING:
    from resolwe.flow.models import Collection, Entity

from .base import VERSION_NUMBER_BITS, AuditModel

VALIDATOR_LENGTH = 128
NAME_LENGTH = 128
LABEL_LENGTH = 128
DESCRIPTION_LENGTH = 256


class AnnotationType(Enum):
    """Supported annotation types."""

    DATE = "DATE"
    DECIMAL = "DECIMAL"
    INTEGER = "INTEGER"
    STRING = "STRING"


class AnnotationValueValidator:
    """Class that validates the annotation value.

    It uses plugins (classes) to validate the annotation of the given type.
    """

    def __init__(self):
        """Initialize."""
        self._validators: Mapping[
            AnnotationType, MutableSequence["AnnotationFieldBaseValidator"]
        ] = defaultdict(list)

    def _add_validator(self, validator: "AnnotationFieldBaseValidator"):
        """Add the given validator."""
        for field_type in validator.annotation_field_types:
            self._validators[field_type].append(validator)

    def validate(
        self, annotation_value: "AnnotationValue", raise_exception: bool = False
    ) -> Sequence[ValidationError]:
        """Validate the given AnnotationValue object.

        All validation errors are grouped together.

        :raises ValidationError: if annotation fails and raise_exception is True. The
            actual errors are nested inside outer ValidationError.
        """
        errors = []
        validators = self._validators[annotation_value.field.annotation_type]
        for validator in validators:
            try:
                validator.validate(annotation_value.value, annotation_value.field)
            except ValidationError as error:
                errors.append(error)
        if errors and raise_exception:
            raise ValidationError(errors)
        return errors


annotation_value_validator = AnnotationValueValidator()


class AnnotationFieldBaseValidator(metaclass=abc.ABCMeta):
    """Base class for annotation class validation."""

    annotation_field_types: Iterable[AnnotationType]

    @classmethod
    def __init_subclass__(cls: Type["AnnotationFieldBaseValidator"], **kwargs):
        """Register instance of the class with the validator."""
        super().__init_subclass__(**kwargs)
        annotation_value_validator._add_validator(cls())

    def validate(self, value: Any, annotation_field: "AnnotationField"):
        """Validate the value.

        :raises ValidatiorError: on validation failure.
        """
        raise NotImplementedError("Subclass should implement the 'validate' method.")


class AnnotationFieldStringValidator(AnnotationFieldBaseValidator):
    """Validates the string field."""

    annotation_field_types = (AnnotationType.STRING,)

    def validate(self, value: Any, annotation_field: "AnnotationField"):
        """Validate value to be of type str."""
        if not isinstance(value, str):
            raise ValidationError(
                f"The value '{value}' is not of the expected type 'str'."
            )


class AnnotationFieldIntegerValidator(AnnotationFieldBaseValidator):
    """Validates the integer field."""

    annotation_field_types = (AnnotationType.INTEGER,)

    def validate(self, value: Any, annotation_field: "AnnotationField"):
        """Validate value to be of type int."""
        if not isinstance(value, int):
            raise ValidationError(
                f"The value '{value}' is not of the expected type 'int'."
            )


class AnnotationFieldDecimalValidator(AnnotationFieldBaseValidator):
    """Validates the decimal field."""

    annotation_field_types = (AnnotationType.DECIMAL,)

    def validate(self, value: Any, annotation_field: "AnnotationField"):
        """Validate value to be of type int or float."""
        if not isinstance(value, (int, float)):
            raise ValidationError(
                f"The value '{value}' is not of the expected types 'int' or 'float'."
            )


class AnnotationFieldDateValidator(AnnotationFieldBaseValidator):
    """Validates the date field."""

    annotation_field_types = (AnnotationType.DATE,)

    def validate(self, value: Any, annotation_field: "AnnotationField"):
        """Validate value represents a date in the format YYYY-MM-DD.

        The month and day part can contain a single character.
        """
        if not isinstance(value, str):
            raise ValidationError(
                f"The value '{value}' is not of the expected type 'str'."
            )
        try:
            datetime.datetime.strptime(value, "%Y-%m-%d")
        except ValueError:
            raise ValidationError(
                f"Value {value} has incorrect format, use YYYY-MM-DD."
            )


class AnnotationFieldRegexValidator(AnnotationFieldBaseValidator):
    """Validates regex for the annotation value."""

    annotation_field_types = (
        AnnotationType.INTEGER,
        AnnotationType.DATE,
        AnnotationType.DECIMAL,
        AnnotationType.STRING,
    )

    def validate(self, value: Any, annotation_field: "AnnotationField"):
        """Validate that the value matches regex defined on the field.

        The value is converted to string before matching.
        """
        # Validate the regex.
        if annotation_field.validator_regex is not None:
            regex = annotation_field.validator_regex
            value = str(value)
            if re.search(regex, value) is None:
                raise ValidationError(
                    f"The value '{value}' for the field '{annotation_field.pk}' does "
                    f"not match the regex '{regex}'."
                )


class AnnotationFieldVocabularyValidator(AnnotationFieldBaseValidator):
    """Validates that the value matches vocabulary."""

    annotation_field_types = (
        AnnotationType.INTEGER,
        AnnotationType.DATE,
        AnnotationType.DECIMAL,
        AnnotationType.STRING,
    )

    def validate(self, value: Any, annotation_field: "AnnotationField"):
        """Validate that the value matches vocabulary on the field."""
        if (
            annotation_field.vocabulary is not None
            and value not in annotation_field.vocabulary
        ):
            raise ValidationError(
                f"The value '{value}' is not valid for the field {annotation_field}."
            )


class AbstractAnnotationGroup(models.Model):
    """Abstract base class for annotation groups."""

    #: the name of the annotation group
    name = models.CharField(max_length=NAME_LENGTH)

    #: the label of the annotation group
    label = models.CharField(max_length=LABEL_LENGTH)

    #: the sorting order among annotation groups
    sort_order = models.PositiveSmallIntegerField()

    class Meta:
        """Make this class abstract."""

        abstract = True


class AnnotationGroup(AbstractAnnotationGroup):
    """Group of annotation fields."""

    class Meta:
        """Set the default ordering."""

        ordering = ["sort_order"]

    def __str__(self) -> str:
        """Return user-friendly string representation."""
        return f"{self.name}"


class AnnotationFieldManager(
    BaseManager["AnnotationField", models.QuerySet["AnnotationField"]]
):
    """Annotation field manager.

    Return only the latest version of each field.
    """

    partition_fields = ["name", "group"]


class AbstractAnnotationField(models.Model):
    """Abstract base class for annotation fields."""

    #: the name of the annotation fields
    name = models.CharField(max_length=NAME_LENGTH)

    #: user visible field name
    label = models.CharField(max_length=LABEL_LENGTH)

    #: user visible field description
    description = models.CharField(max_length=DESCRIPTION_LENGTH)

    #: the type of the annotation field
    type = models.CharField(max_length=16)

    #: the annotation group this field belongs to
    group = models.ForeignKey(
        AnnotationGroup, on_delete=models.CASCADE, related_name="fields"
    )

    #: the sorting order among annotation fields
    sort_order = models.PositiveSmallIntegerField()

    #: optional regular expression for validation
    validator_regex = models.CharField(
        max_length=VALIDATOR_LENGTH, null=True, blank=True
    )

    #: optional map of valid values to labels
    vocabulary = models.JSONField(null=True, blank=True)

    #: is this field required
    required = models.BooleanField(default=False)

    #: field version
    version = VersionField(number_bits=VERSION_NUMBER_BITS, default="0.0.0")

    class Meta:
        """Make this class abstract."""

        abstract = True

    def __init__(self, *args, **kwargs):
        """Store original vocabulary to private variable."""
        super().__init__(*args, **kwargs)
        self._original_vocabulary = self.vocabulary
        self._original_type = self.type

    @staticmethod
    def add_to_collection(source: "Collection", destination: "Collection"):
        """Add fields from the source to the destination collection."""
        destination.annotation_fields.add(*source.annotation_fields.all())

    @property
    def annotation_type(self) -> AnnotationType:
        """Get the field type as enum."""
        return AnnotationType(self.type.upper())

    @staticmethod
    def group_field_from_path(path: str) -> List[str]:
        """Return the group and field name from path."""
        match = re.fullmatch(
            r"(?P<group>[a-zA-Z][\w ]+)\.(?P<field>[a-zA-Z][\w ]+)", path, re.ASCII
        )
        if match is None:
            raise ValidationError(f"Invalid path '{path}'.")
        return [match["group"], match["field"]]

    @classmethod
    def id_from_path(cls, path: str) -> Optional[int]:
        """Get the field id from the field path."""
        group_name, field_name = cls.group_field_from_path(path)
        return (
            cls.objects.filter(group__name=group_name, name=field_name)
            .values_list("id", flat=True)
            .first()
        )

    @classmethod
    def field_from_path(cls, path: str) -> "AnnotationField":
        """Get the field id from the field path.

        :raises ValidationError: when field does not exist.
        """
        group_name, field_name = cls.group_field_from_path(path)
        field = cls.objects.filter(group__name=group_name, name=field_name).first()
        if not field:
            raise ValidationError(f"Field '{path}' does not exist.")
        return field

    def save(self, *args, **kwargs):
        """Recompute the labels for annotation values if vocabulary changes.

        :raises ValidationError: when vocabulary changes so that annotation values are no
            longer valid.
        """
        super().save(*args, **kwargs)
        if self.vocabulary != self._original_vocabulary:
            updated_fields = []
            for annotation_value_field in self.values.all():
                annotation_value_field.recompute_label()
                annotation_value_field.validate()
                updated_fields.append(annotation_value_field)
            AnnotationValue.all_objects.bulk_update(updated_fields, ["_value"])
            self._original_vocabulary = self.vocabulary
        if self.type != self._original_type:
            self.revalidate_values()
            self._original_type = self.type

    def revalidate_values(self):
        """Revalidate all annotation values.

        :raises ValidationError: when validation fails.
        """
        for annotation_value in self.values.iterator():
            annotation_value.validate()

    def label_by_value(self, label: str) -> str:
        """Get the value by label.

        When no value is found the label is returned.
        """
        if self.vocabulary is not None:
            for vocabulary_value, vocabulary_label in self.vocabulary.items():
                if label == vocabulary_label:
                    return vocabulary_value
        return label


class AnnotationField(AbstractAnnotationField):
    """Annotation field."""

    #: the latest version of the models
    objects = AnnotationFieldManager()

    #: all models
    all_objects = models.Manager()

    class Meta:
        """Set the constraints and the default ordering."""

        constraints = [
            # Accept only supported annotation types.
            models.constraints.CheckConstraint(
                check=models.Q(type__in=[e.value for e in AnnotationType]),
                name="annotation_type",
            ),
            models.constraints.UniqueConstraint(
                fields=["name", "group", "version"],
                name="uniquetogether_name_group_version",
            ),
        ]
        ordering = ["group__sort_order", "sort_order"]

    def __str__(self) -> str:
        """Return user-friendly string representation."""
        return f"{self.group.name}.{self.name}"


class AnnotationPreset(PermissionObject, BaseModel):
    """The named set of annotation fields.

    The presets have permissions.
    """

    #: the fields belonging to this preset
    fields = models.ManyToManyField(AnnotationField, related_name="presets")

    def __str__(self):
        """Return the string representation."""
        return self.name

    def can_set_permission(self) -> bool:
        """Return whether permissions can be set."""
        return True

    class Meta:
        """Override parent meta."""


class AnnotationValueManager(BaseManager["AnnotationValue", PermissionQuerySet]):
    """Annotation value manager.

    Return only the latest version of each field.
    """

    partition_fields = ["field", "entity"]
    version_field = "created"
    QuerySet = PermissionQuerySet

    def remove_delete_markers(self) -> PermissionQuerySet:
        """Remote delete markers from the queryset.

        Due to the nature of the queryset use the method only on small querysets.
        """
        return self.filter(pk__in=list(self.values_list("pk", flat=True))).exclude(
            _value__isnull=True
        )


def _slug_for_annotation_value(instance: "AnnotationValue") -> str:
    """Generate the slug for the annotation value."""
    return f"{instance.entity.slug}-{instance.field.group.name}-{instance.field.name}"


class AbstractAnnotationValue(BaseModel):
    """Abstract base class for annotation values."""

    #: URL slug
    slug = ResolweSlugField(
        populate_from=_slug_for_annotation_value,
        unique_with="version",
        max_length=MAX_SLUG_LENGTH,
    )

    #: the entity this field belongs to
    entity: "Entity" = models.ForeignKey(
        "Entity", related_name="annotations", on_delete=models.CASCADE
    )  # type: ignore

    #: the field this field belongs to
    field = models.ForeignKey(
        AnnotationField, related_name="values", on_delete=models.PROTECT
    )

    #: value is stored under key 'value' in the json field to simplify lookups
    #: the value None is used as delete marker
    _value: Any = models.JSONField(default=dict, null=True)

    class Meta:
        """Make this class abstract."""

        abstract = True

    def __init__(self, *args, **kwargs):
        """Allow us to set the 'value' in the constructor.

        This is a little tricky, as value can be passed in args or kwargs. We only
        intercept the value when given as kwarg, since args are used by Django when
        constructing class from database value.
        """
        if "value" in kwargs:
            kwargs["_value"] = {"value": kwargs.pop("value")}
        super().__init__(*args, **kwargs)
        if not self.delete_marker and "label" not in self._value:
            self.recompute_label()

    @property
    def delete_marker(self) -> bool:
        """Return True if this value represents a delete marker."""
        return self._value is None

    @property
    def value(self) -> Optional[str | int | float | datetime.date]:
        """Get the actual value or None if object is delete marker."""
        return None if self.delete_marker else self._value["value"]

    @value.setter
    def value(self, value: str | int | float | datetime.date):
        """Set the value.

        The label is recomputed when the value is set.
        """
        self._value["value"] = value
        self.recompute_label()

    @property
    def _computed_label(self) -> Any:
        """Return computed label from value.

        :raises KeyError: when the value is not in the vocabulary.
        """
        if self.field.vocabulary is None:
            return self.value
        else:
            # Always return a value even if it is not in the vocabulary. The validation
            # step will take care of the wrong values.
            return self.field.vocabulary.get(self.value, self.value)

    @property
    def label(self) -> Any:
        """Return the cached label."""
        return self._value["label"]

    def validate(self):
        """Validate the given value.

        The validation is always done in full, all errors are gathered and sent
        in the response.

        :raises ValidationError: when the validation fails.
        """
        if not self.delete_marker:
            annotation_value_validator.validate(self, raise_exception=True)

    def save(self, *args, **kwargs):
        """Save the annotation value after validation has passed."""
        # The value `None` is used as delete marker.
        if not self.delete_marker:
            annotation_value_validator.validate(self, raise_exception=True)
            # Make sure the label is always set.
            if "label" not in self._value:
                self.recompute_label()
        super().save(*args, **kwargs)

    def recompute_label(self):
        """Recompute label from value and set it to the model instance."""
        self._value["label"] = self._computed_label

    @staticmethod
    def from_path(entity_id: int, path: str) -> Optional["AnnotationValue"]:
        """Get the annotation value from the path."""
        field_id = AnnotationField.id_from_path(path)
        return AnnotationValue.objects.filter(
            entity_id=entity_id, field_id=field_id
        ).first()

    @classmethod
    def permission_proxy(cls) -> str:
        """Return the permission group path."""
        return "entity"

    def in_container(self) -> bool:
        """Return if object lies in a container."""
        return True


class AnnotationValue(AbstractAnnotationValue, PermissionInterface, AuditModel):
    """The value of the annotation."""

    #: the latest version of the models
    objects = AnnotationValueManager()  # type: ignore

    #: all models
    all_objects: models.Manager["AnnotationValue"] = PermissionQuerySet.as_manager()

    class Meta:
        """Set the unique constraints."""

        constraints = [
            models.constraints.UniqueConstraint(
                fields=["entity", "field", "created"],
                name="uniquetogether_entity_field_created",
            ),
        ]
        ordering = ["field__group__sort_order", "field__sort_order"]

    def __str__(self) -> str:
        """Return user-friendly string representation."""
        if self.delete_marker:
            return "Delete marker"
        else:
            return f"{self.label}"



# pylint: disable=missing-docstring
from resolwe.flow.models import Process
from resolwe.test import TestCase


class ManagerTest(TestCase):
    def setUp(self):
        super().setUp()

        self.data = {
            "name": "Test Process",
            "contributor": self.contributor,
            "type": "data:test",
            "version": 1,
        }

    def test_slug(self):
        p = Process.objects.create(**self.data)
        self.assertEqual(p.slug, "test-process")

        self.data["version"] = 2
        p = Process.objects.create(**self.data)
        self.assertEqual(p.slug, "test-process")

        p = Process.objects.create(**self.data)
        self.assertEqual(p.slug, "test-process-2")



"""Storage application views."""

import json
import logging
import re
import tempfile
from datetime import datetime
from distutils.util import strtobool
from functools import lru_cache
from pathlib import Path
from typing import Dict, List, Tuple, Union
from urllib.parse import urlencode, urlparse

import pytz
from bs4 import BeautifulSoup
from django.core.exceptions import ImproperlyConfigured, PermissionDenied
from django.http import HttpRequest, HttpResponse, JsonResponse
from django.shortcuts import redirect
from django.urls import reverse
from django.utils import timezone
from django.utils.decorators import method_decorator
from django.views import View
from django.views.decorators.csrf import csrf_exempt
from rest_framework import serializers
from rest_framework.response import Response
from rest_framework.views import APIView

from resolwe.flow.models import Data
from resolwe.flow.models.collection import Collection
from resolwe.permissions.utils import get_user
from resolwe.storage.connectors import ConnectorType, connectors
from resolwe.storage.connectors.baseconnector import BaseStorageConnector
from resolwe.storage.models import FileStorage, ReferencedPath
from resolwe.test.utils import ignore_in_tests

logger = logging.getLogger(__name__)


class UploadConfigSerializer(serializers.Serializer):
    """Serializer for upload configuration."""

    type = serializers.CharField()
    config = serializers.DictField()


class RFC3339DateTimeField(serializers.DateTimeField):
    """DateTime field that serializes to RFC3339 format."""

    format = "%Y-%m-%dT%H:%M:%SZ"


class UploadCredentialsSerializer(serializers.Serializer):
    """Serializer for upload credentials."""

    AccessKeyId = serializers.CharField()
    SecretAccessKey = serializers.CharField()
    Token = serializers.CharField()
    Expiration = RFC3339DateTimeField()


class UploadCredentials(APIView):
    """Get the upload credentials."""

    # Used to generate the API schema.
    serializer_class = UploadCredentialsSerializer

    def get(self, request, *args, **kwargs):
        """Return the upload credentials."""
        try:
            upload_connector = connectors.for_storage("upload")[0]
        except Exception:
            message = "Upload connector could not be determined."
            logger.exception(message)
            raise ImproperlyConfigured(message)

        if upload_connector.CONNECTOR_TYPE != ConnectorType.S3:
            raise ImproperlyConfigured(
                "Credentials endpoint only supports S3 connector."
            )
        prefix = str(get_user(request.user).id)
        credentials = upload_connector.temporary_credentials(prefix)["credentials"]
        # The credentials are returned in the format expected by the AWS SDK, except
        # the key "SessionToken", which must be transformed to "Token".
        credentials["Token"] = credentials.pop("SessionToken")
        return Response(self.serializer_class(credentials).data)


class UploadConfig(APIView):
    """Get the upload configuration."""

    serializer_class = UploadConfigSerializer

    def get(self, request, *args, **kwargs):
        """Return the JSON representing the upload configuration.

        The returning object is JSON representation of the dictionary with the
        following fields:

        - type: the type of upload connector. Currently we
          support 'LOCAL', and 'S3' connector types. Upload through server is
          always supported.

        - credentials: the dictionary representing the set of credentials that
          are used to upload data. This dictionary is specific to connector
          type and may be empty.
        """
        try:
            upload_connector = connectors.for_storage("upload")[0]
            prefix = str(get_user(request.user).id)
            response = {
                "type": upload_connector.CONNECTOR_TYPE.name,
                "config": upload_connector.temporary_credentials(prefix),
            }
        except Exception:
            message = "Upload connector could not be determined."
            logger.exception(message)
            raise ImproperlyConfigured(message)

        serializer = self.serializer_class(response)
        return Response(serializer.data)


class DataBrowseView(View):
    """Emulate nginx JSON directory view of the data directory.

    Use information in database to construct JSON response that emulates nginx
    autoindex module with autoindex_format set to json. See
    http://nginx.org/en/docs/http/ngx_http_autoindex_module.html
    for details.
    """

    def _get_mtime(self, entry: datetime) -> str:
        """Convert datetime entry to nginx autoindex compatible string.

        The timezone returned by autoindex module was set to GMT so we convert
        from our timetone to GMT.
        """
        gmt = timezone.localtime(entry, pytz.timezone("GMT"))
        return gmt.strftime("%a, %d %b %Y %H:%M:%S %Z")

    def _path_to_dict(
        self, referenced_path: ReferencedPath, base_path: Union[str, Path]
    ) -> Dict[str, Union[str, int]]:
        """Convert ReferencedPath to dictionary emulating nginx response."""
        assert referenced_path.file_storage is not None
        data = {
            "name": Path(referenced_path.path).relative_to(base_path).as_posix(),
            "type": "file",
            "mtime": self._get_mtime(referenced_path.file_storage.created),
            "size": referenced_path.size,
            "md5": referenced_path.md5,
        }
        if referenced_path.path.endswith("/"):
            data.pop("size")
            data["type"] = "directory"
        return data

    def _get_datum(self, data_id: int) -> Data:
        """Get datum with given id and check it's permissions."""
        # Check permissions and load requested data object.
        data = Data.objects.filter(pk=data_id).filter_for_user(self.request.user)
        # Send response with status 403 when requested data object does not exists.
        if not data.exists():
            raise PermissionDenied()

        return data.get()

    def _resolve_dir(
        self, relative_path: Path, file_storage: FileStorage
    ) -> List[Dict[str, Union[str, int]]]:
        """Resolve directory."""
        # Empty path evaluates to "."
        regex_path = relative_path.as_posix() + "/" if relative_path != Path() else ""
        # Show only entries in this directory.
        regex = "^{}[^/]+/?$".format(regex_path)
        return [
            self._path_to_dict(path, regex_path)
            for path in file_storage.files.filter(path__regex=regex)
        ]

    def _resolve_file(self, relative_path: Path, file_storage: FileStorage) -> str:
        """Resolve file URI and return signed url.

        :raises PermissionDenied: if presigned URL could not be generated.
        """
        # Redirect to the resource.
        force_download = strtobool(self.request.GET.get("force_download", "false"))
        default_storage_location = file_storage.default_storage_location
        assert default_storage_location is not None
        presigned_url = default_storage_location.connector.presigned_url(
            file_storage.subpath / relative_path, force_download=force_download
        )
        if presigned_url is None:
            raise PermissionDenied()

        return presigned_url

    def _get_response(
        self, datum: Data, relative_path: Path
    ) -> Tuple[Union[str, list], bool]:
        """Return a (response, is_file) tuple.

        :raises PermissionDenied: when path can not be resolved.
        """
        file_storage: FileStorage = datum.location

        if (
            relative_path == Path()
            or file_storage.files.filter(path=relative_path.as_posix() + "/").exists()
        ):
            # If directory
            return (self._resolve_dir(relative_path, file_storage), False)
        elif file_storage.files.filter(path=relative_path.as_posix()).exists():
            # If file
            return (self._resolve_file(relative_path, file_storage), True)

        raise PermissionDenied()

    def get(
        self, request: HttpRequest, *args: Union[str, int], **kwargs: Union[str, int]
    ) -> HttpResponse:
        """Giver URL either redirect to resource or serve a list of resources.

        If URL points to a file then redirect to a default storage location.
        If URL points to a directory then send a JSON encoded list of entires
        in this directory emulating nginx autoindex module.

        :param request: the given request. When the user requests URL
            http://platform-centos.genialis.local/data/1/..., then the value of
            parameter should be 1/... .
        """

        data_id = int(kwargs["data_id"])
        relative_path = str(kwargs.get("uri", ""))
        if Path(urlparse(relative_path).path).suffix in [".html", ".htm"]:
            redirect_url = reverse("HTMLPreprocessorView", args=args, kwargs=kwargs)
            arguments = urlencode(request.GET)
            if arguments:
                redirect_url = f"{redirect_url}?{arguments}"
            return redirect(redirect_url)

        datum = self._get_datum(data_id)
        response, is_file = self._get_response(datum, Path(relative_path))
        if is_file:
            return redirect(response)
        elif response:
            return JsonResponse(response, safe=False)
        else:
            raise PermissionDenied()


class HTMLPreprocessorView(DataBrowseView):
    """Serve HTML files.

    Pull the file to the server, pre-sign all the links in the file (including
    CSS and Javascript) and serve the modified file.
    """

    def get(
        self, request: HttpRequest, *args: Union[str, int], **kwargs: Union[str, int]
    ) -> HttpResponse:
        """Serve transformed HTML."""

        def _process_url(
            url: str, from_connector: BaseStorageConnector, subpath: Path
        ) -> str:
            """Presign URL if necessary."""
            parsed_url = urlparse(url)
            # We have to process link when it is relative (no scheme and no
            # netloc is given). Is should also not be anchor link (no path).
            if parsed_url.scheme or parsed_url.netloc or not parsed_url.path:
                return url

            # HTML files should be redirected to HTMLProcessor view.
            if Path(parsed_url.path).suffix in [".htm", ".html"]:
                return reverse("HTMLPreprocessorView", args=args, kwargs=kwargs)
            else:
                presigned_url = from_connector.presigned_url(
                    subpath / url, expiration=3600
                )
                if presigned_url is None:
                    raise RuntimeError(
                        "Unable to presign URL {url} for data object with id {subpath}."
                    )
                return presigned_url

        data_id = int(kwargs["data_id"])
        relative_path = Path(str(kwargs.get("uri", "")))
        assert relative_path.suffix in [
            ".html",
            ".htm",
        ], f"Only HTML files can be served, {relative_path} requested."
        datum = self._get_datum(data_id)
        subpath = Path(datum.location.subpath)
        from_connector = datum.location.default_storage_location.connector
        # Temporary file is opened with mode 'w+b' by default.
        with tempfile.SpooledTemporaryFile() as stream:
            from_connector.get(subpath / relative_path, stream)
            stream.seek(0)
            soup = BeautifulSoup(stream)
            tags_attributes = {
                "a": "href",
                "link": "href",
                "script": "src",
                "img": "src",
            }
            for tag in soup.findAll(list(tags_attributes.keys())):
                attribute = tags_attributes[tag.name]
                if attribute in tag.attrs:
                    tag[attribute] = _process_url(
                        tag[attribute], from_connector, subpath
                    )
            result = str(soup)
        response = HttpResponse(result)
        if strtobool(self.request.GET.get("force_download", "false")):
            response["Content-Disposition"] = "attachment"
        return response


@method_decorator(csrf_exempt, name="dispatch")
class UriResolverView(DataBrowseView):
    """Get resolved and signed URLs for a given list of URIs."""

    def get(
        self, request: HttpRequest, *args: Union[str, int], **kwargs: Union[str, int]
    ) -> HttpResponse:
        """
        Return empty response.

        GET Method should not be disabled, but since it is implemented in
        superclass, it is overwritten.
        """
        return HttpResponse("")

    def post(
        self, request: HttpRequest, *args: Union[str, int], **kwargs: Union[str, int]
    ) -> HttpResponse:
        """Get resolved and signed URLs for a given list of URIs."""

        content = {}
        try:
            content = json.loads(request.read())
        except json.decoder.JSONDecodeError:
            pass

        response_data = {}
        for uri in content.get("uris", []):
            match = re.match(r"(\d+)/(.+)", uri)
            if not match:
                response_data[uri] = ""
                continue
            data_id = int(match.group(1))
            relative_path = Path(match.group(2))
            datum = self._get_datum(data_id)
            try:
                response_data[uri] = self._get_response(datum, relative_path)[0]
            except PermissionDenied:
                response_data[uri] = ""

        return JsonResponse(response_data)

    def _get_datum(self, data_id: int) -> Data:
        """
        Get datum with given id and check it's permissions.

        Speedup: Since checking permissions if quite slow, this method
        can become a real bottleneck if there are > 100 Data object to
        check. However, if they are from the same collection it is easy
        to check, just confirm that the user has view permission on
        collection. That can be checked only once and than cached.
        Except in tests, where colletion ids may be reused.
        """

        @ignore_in_tests(lru_cache(maxsize=1))
        def has_view_permission(collection_id: int) -> bool:
            collection = Collection.objects.filter(pk=collection_id).filter_for_user(
                self.request.user
            )
            return collection.exists()

        datum = Data.objects.get(pk=data_id)

        if datum.collection and has_view_permission(datum.collection.id):
            return datum

        return super()._get_datum(data_id)



# Generated by Django 4.2.16 on 2024-11-21 16:23

import django.db.models.manager
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0031_remove_annotationvalue_uniquetogether_entity_field_and_more"),
    ]

    operations = [
        migrations.AlterModelManagers(
            name="process",
            managers=[
                ("unversioned_objects", django.db.models.manager.Manager()),
            ],
        ),
    ]



"""Overrides for Django settings in tests."""

import copy

from django.conf import settings

from resolwe.storage import settings as storage_settings


def prepare_executor_settings():
    """Override FLOW_EXECUTOR settings with settings specified in FLOW_EXECUTOR['TEST']."""
    global FLOW_EXECUTOR_SETTINGS
    FLOW_EXECUTOR_SETTINGS = copy.deepcopy(getattr(settings, "FLOW_EXECUTOR", {}))
    TEST_SETTINGS_OVERRIDES = getattr(settings, "FLOW_EXECUTOR", {}).get("TEST", {})
    FLOW_EXECUTOR_SETTINGS.update(TEST_SETTINGS_OVERRIDES)


def prepare_manager_settings():
    """Override FLOW_MANAGER settings with settings specified in FLOW_MANAGER['TEST']."""
    global FLOW_MANAGER_SETTINGS
    FLOW_MANAGER_SETTINGS = copy.deepcopy(getattr(settings, "FLOW_MANAGER", {}))
    TEST_SETTINGS_OVERRIDES = getattr(settings, "FLOW_MANAGER", {}).get("TEST", {})
    FLOW_MANAGER_SETTINGS.update(TEST_SETTINGS_OVERRIDES)


def prepare_storage_settings():
    """Prepare the storage connectors and volumes settings."""
    global STORAGE_CONNECTORS, FLOW_VOLUMES
    STORAGE_CONNECTORS = copy.deepcopy(storage_settings.STORAGE_CONNECTORS)
    for connector_settings in STORAGE_CONNECTORS.values():
        connector_settings["config"].update(connector_settings.get("test", {}))

    FLOW_VOLUMES = copy.deepcopy(storage_settings.FLOW_VOLUMES)
    for volume in FLOW_VOLUMES.values():
        volume["config"].update(volume.get("test", {}))


# Prepare the settings overrides.
prepare_executor_settings()
prepare_manager_settings()
prepare_storage_settings()



""".. Ignore pydocstyle D400.

==============
Flow Utilities
==============

.. automodule:: resolwe.flow.utils.exceptions
   :members:

.. automodule:: resolwe.flow.utils.stats
   :members:

"""

import functools
import hashlib
import json
from pathlib import Path
from typing import Dict

from django.apps import apps
from django.conf import settings
from django.db import models
from versionfield.version import Version

from .decorators import retry
from .iterators import iterate_dict, iterate_fields, iterate_schema
from .singleton import Singleton

__all__ = (
    "dict_dot",
    "get_apps_tools",
    "get_data_checksum",
    "iterate_dict",
    "iterate_fields",
    "iterate_schema",
    "retry",
    "Singleton",
)


def get_data_checksum(proc_input: Dict, proc_slug: str, proc_version: Version) -> str:
    """Compute checksum of processor inputs, name and version."""
    checksum = hashlib.sha256()
    checksum.update(json.dumps(proc_input, sort_keys=True).encode("utf-8"))
    checksum.update(proc_slug.encode("utf-8"))
    checksum.update(str(proc_version).encode("utf-8"))
    return checksum.hexdigest()


def dict_dot(d: Dict, k, val=None, default=None):
    """Get or set value using a dot-notation key in a multilevel dict."""
    if val is None and k == "":
        return d

    def set_default(dict_or_model, key, default_value):
        """Set default field value."""
        if isinstance(dict_or_model, models.Model):
            if not hasattr(dict_or_model, key):
                setattr(dict_or_model, key, default_value)

            return getattr(dict_or_model, key)
        else:
            return dict_or_model.setdefault(key, default_value)

    def get_item(dict_or_model, key):
        """Get field value."""
        if isinstance(dict_or_model, models.Model):
            return getattr(dict_or_model, key)
        else:
            return dict_or_model[key]

    def set_item(dict_or_model, key, value):
        """Set field value."""
        if isinstance(dict_or_model, models.Model):
            setattr(dict_or_model, key, value)
        else:
            dict_or_model[key] = value

    if val is None and callable(default):
        # Get value, default for missing
        return functools.reduce(
            lambda a, b: set_default(a, b, default()), k.split("."), d
        )

    elif val is None:
        # Get value, error on missing
        return functools.reduce(get_item, k.split("."), d)

    else:
        # Set value
        try:
            k, k_last = k.rsplit(".", 1)
            set_item(dict_dot(d, k, default=dict), k_last, val)
        except ValueError:
            set_item(d, k, val)
        return val


def get_apps_tools() -> Dict[str, Path]:
    """Get applications' tools and their paths.

    Return a dict with application names as keys and paths to tools'
    directories as values. Applications without tools are omitted.

    :raises KeyError: when RESOLWE_CUSTOM_TOOLS_PATHS settings is no a list.
    """
    tools_paths = {}

    for app_config in apps.get_app_configs():
        proc_path = Path(app_config.path) / "tools"
        if proc_path.is_dir():
            tools_paths[app_config.name] = proc_path

    custom_tools_paths = getattr(settings, "RESOLWE_CUSTOM_TOOLS_PATHS", [])
    if not isinstance(custom_tools_paths, list):
        raise KeyError("`RESOLWE_CUSTOM_TOOLS_PATHS` setting must be a list.")

    for seq, custom_path in enumerate(custom_tools_paths):
        custom_key = "_custom_{}".format(seq)
        tools_paths[custom_key] = Path(custom_path)

    return tools_paths



"""Resolwe history model.

It is used to track changes in objects.
"""

from typing import Any, Iterable, Type

from django.contrib.postgres.fields import DateTimeRangeField
from django.db import models
from django.db.backends.postgresql.psycopg_any import DateTimeTZRange

from .base import MAX_NAME_LENGTH, MAX_SLUG_LENGTH
from .collection import Collection
from .data import Data
from .history_manager import HistoryMixin, history_manager


class BaseHistory(models.Model):
    """BaseHistory object."""

    class Meta:
        """Make this model abstract."""

        abstract = True

    #: When this object is valid for
    valid = DateTimeRangeField()

    #: When the object was deleted
    deleted = models.DateTimeField(blank=True, null=True)


class DataHistory(BaseHistory):
    """Track Data history."""

    #: Pointer to the data object, can be null
    datum = models.ForeignKey(
        Data, on_delete=models.SET_NULL, related_name="history", null=True
    )


class CollectionHistory(BaseHistory):
    """Track Collection history."""

    #: Pointer to the collection object, can be null
    datum = models.ForeignKey(
        Collection, on_delete=models.SET_NULL, related_name="history", null=True
    )


class TrackChange(models.Model):
    """Base class to track the changes in the model."""

    class Meta:
        """TrackChange meta options."""

        abstract = True
        ordering = ["timestamp"]
        indexes = [
            models.Index(name="idx_%(class)s_time", fields=["timestamp"]),
        ]

    # Override in the child class.
    # The itarable of tuple (model, field_name) this tracker can be applied to.
    applicable_to: Iterable[tuple[Type[models.Model], str]]

    def __init_subclass__(cls: Type["TrackChange"], **kwargs):
        """Register instance of the class as plugin."""
        super().__init_subclass__(**kwargs)
        history_manager.add_tracker(cls)

    #: Defined to keep mypy from reporting errors (objects do not exist since class is
    #: abstract).
    objects: models.Manager

    #: Timestamp of the change
    timestamp = models.DateTimeField(auto_now_add=True)

    # Here for mypy to detect all objects of type TrackChange must have value.
    value: Any

    def get_value(self, field_name: str) -> Any:
        """Return the value used to track changes."""
        return self.value

    @staticmethod
    def preprocess_value(instance: models.Model, field_name: str) -> Any:
        """Preprocess the value before saving."""
        return getattr(instance, field_name)

    @classmethod
    def process_change(cls, instance: HistoryMixin, field_name: str):
        """Create new track change object."""
        value = cls.preprocess_value(instance, field_name)
        cls.objects.create(history=instance.get_history_object(), value=value)


class DataSlugChange(TrackChange):
    """Track data slug changes."""

    applicable_to = ((Data, "slug"),)

    #: New value.
    value = models.CharField(max_length=MAX_SLUG_LENGTH)

    #: Reference to the history object
    history = models.ForeignKey(
        DataHistory, on_delete=models.PROTECT, related_name="%(class)s"
    )


class CollectionSlugChange(TrackChange):
    """Track collection slug changes."""

    applicable_to = ((Collection, "slug"),)

    #: New value.
    value = models.CharField(max_length=MAX_SLUG_LENGTH)

    #: Reference to the history object
    history = models.ForeignKey(
        CollectionHistory, on_delete=models.PROTECT, related_name="%(class)s"
    )


class DataNameChange(TrackChange):
    """Track data name changes."""

    applicable_to = ((Data, "name"),)

    #: New value.
    value = models.CharField(max_length=MAX_NAME_LENGTH)

    #: Reference to the history object
    history = models.ForeignKey(
        DataHistory, on_delete=models.PROTECT, related_name="%(class)s"
    )


class CollectionNameChange(TrackChange):
    """Track collection name changes."""

    applicable_to = ((Collection, "name"),)

    #: New value.
    value = models.CharField(max_length=MAX_NAME_LENGTH)

    #: Reference to the history object
    history = models.ForeignKey(
        CollectionHistory, on_delete=models.PROTECT, related_name="%(class)s"
    )


class CollectionChange(TrackChange):
    """Track collection changes."""

    applicable_to = ((Data, "collection"),)

    #: New value.
    value = models.ForeignKey(
        CollectionHistory,
        on_delete=models.PROTECT,
        related_name="collection_changes",
        null=True,
    )

    #: Reference to the history object
    history = models.ForeignKey(
        DataHistory, on_delete=models.PROTECT, related_name="%(class)s"
    )

    @staticmethod
    def preprocess_value(instance: models.Model, field_name: str) -> Any:
        """Preprocess the value before saving."""
        value = getattr(instance, field_name)
        if isinstance(value, Collection):
            return value.get_history_object()
        return value


class SizeChange(TrackChange):
    """Track size changes."""

    applicable_to = ((Data, "size"),)

    #: New value
    value = models.BigIntegerField()

    #: Reference to the history object
    history = models.ForeignKey(
        DataHistory, on_delete=models.PROTECT, related_name="%(class)s"
    )


class ProcessingHistory(TrackChange):
    """Track processing changes."""

    applicable_to = ((Data, "finished"),)

    #: Reference to the history object
    history = models.ForeignKey(
        BaseHistory, on_delete=models.PROTECT, related_name="processing"
    )

    #: Processing interval
    interval = DateTimeRangeField()

    #: Number of CPU allocated to the process
    allocated_cpu = models.IntegerField()

    #: Amount of memory (in bytes) allocated to the process
    allocated_memory = models.BigIntegerField()

    #: Number of CPU on machine running the process
    node_cpu = models.IntegerField()

    #: Amount of memory on machine running the process
    node_memory = models.BigIntegerField()

    #: Reference to the history object
    history = models.ForeignKey(
        DataHistory, on_delete=models.PROTECT, related_name="%(class)s"
    )

    def get_value(self, field_name: str) -> Any:
        """Return the value used to track changes."""
        return self.interval.upper

    @classmethod
    def process_change(cls, instance: HistoryMixin, field_name: str):
        """Preprocess the value before saving."""
        assert isinstance(instance, Data), "The instance must be Data instance."
        assert field_name == "finished", "The field name must be 'finished'."
        started = instance.started
        finished = instance.finished
        history = instance.get_history_object()

        if finished is not None:
            return ProcessingHistory.objects.create(
                interval=DateTimeTZRange(lower=started, upper=finished),
                history=history,
                allocated_cpu=instance.process_cores,
                allocated_memory=instance.process_memory,
                node_cpu=instance.process_cores,
                node_memory=instance.process_memory,
            )
        return cls.objects.last()
